import os

from tornado.web import addslash

import sickchill.update_manager
from sickchill import settings
from sickchill.oldbeard import helpers
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from .index import Config


@Route("/config/backuprestore(/?.*)", name="config:backup")
class ConfigBackupRestore(Config):
    @addslash
    def index(self):
        t = PageTemplate(rh=self, filename="config_backuprestore.mako")

        return t.render(
            submenu=self.ConfigMenu(),
            title=_("Config - Backup/Restore"),
            header=_("Backup/Restore"),
            topmenu="config",
            controller="config",
            action="backupRestore",
        )

    def backup(self):
        backup_dir = self.get_body_argument("backupDirectory", default="")
        final_result = ""

        if backup_dir:
            updater = sickchill.update_manager.UpdateManager()
            target = updater.backup_to_dir(backup_dir)
            if target:
                final_result += f"Successful backup to {target}"
            else:
                final_result += "Backup FAILED"
        else:
            final_result += "You need to choose a folder to save your backup to!"

        final_result += "<br>\n"

        return final_result

    def restore(self):
        backup_file = self.get_body_argument("backupFile", default="")

        final_result = ""

        if backup_file:
            source = backup_file
            target_dir = os.path.join(settings.DATA_DIR, "restore")

            if helpers.restore_config_zip(source, target_dir):
                final_result += "Successfully extracted restore files to " + target_dir
                final_result += "<br>Restart sickchill to complete the restore."
            else:
                final_result += "Restore FAILED"
        else:
            final_result += "You need to select a backup file to restore!"

        final_result += "<br>\n"

        return final_result



import json

import requests
from requests.structures import CaseInsensitiveDict

from sickchill import logger, settings
from sickchill.oldbeard.common import (
    NOTIFY_DOWNLOAD,
    NOTIFY_LOGIN,
    NOTIFY_LOGIN_TEXT,
    NOTIFY_SNATCH,
    NOTIFY_SUBTITLE_DOWNLOAD,
    NOTIFY_UPDATE,
    NOTIFY_UPDATE_TEXT,
    notifyStrings,
)
from sickchill.oldbeard.helpers import make_session


class Notifier(object):
    """
    Use Gotify to send notifications

    https://gotify.net/
    """

    def __init__(self):
        self.session = make_session()

    def test_notify(self, host=None, token=None):
        """
        Send a test notification

        :param host: The host where Gotify is running (incl port)
        :param token: Gotify authorization token
        :returns: the notification
        """
        return self._notify_gotify("Test", "This is a test notification from SickChill", host, token, force=True)

    def _send_gotify_msg(self, title, msg, host=None, token=None):
        """
        Send a Gotify notification

        :param title: The title of the notification to send
        :param msg: The message string to send
        :param host: The host where Gotify is running (incl port)
        :param token: Gotify authorization token
        :return: True if the message succeeded, False otherwise and error
        """
        logger.debug("Gotify in use with authorization token.")
        headers = CaseInsensitiveDict()
        headers["Content-Type"] = "application/json"
        headers["Authorization"] = f"Bearer {token or settings.GOTIFY_AUTHORIZATIONTOKEN}"
        try:
            r = requests.post(
                f"{host}message",
                data=json.dumps(dict(title=title, message=f"{title} : {msg}")),
                headers=headers,
            )
            r.raise_for_status()
        except Exception as error:
            logger.exception(f"Error Sending Gotify message: {error}")
            return False, error
        return True, None

    def notify_snatch(self, ep_name, title=notifyStrings[NOTIFY_SNATCH]):
        """
        Sends a Gotify notification when an episode is snatched

        :param ep_name: The name of the episode snatched
        :param title: The title of the notification to send
        """
        if settings.GOTIFY_NOTIFY_ONSNATCH:
            self._notify_gotify(title, ep_name)

    def notify_download(self, ep_name, title=notifyStrings[NOTIFY_DOWNLOAD]):
        """
        Sends a Gotify notification when an episode is downloaded

        :param ep_name: The name of the episode downloaded
        :param title: The title of the notification to send
        """
        if settings.GOTIFY_NOTIFY_ONDOWNLOAD:
            self._notify_gotify(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=notifyStrings[NOTIFY_SUBTITLE_DOWNLOAD]):
        """
        Sends a Gotify notification when subtitles for an episode are downloaded

        :param ep_name: The name of the episode subtitles were downloaded for
        :param lang: The language of the downloaded subtitles
        :param title: The title of the notification to send
        """
        if settings.GOTIFY_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notify_gotify(title, "{0}: {1}".format(ep_name, lang))

    def notify_update(self, new_version="??"):
        """
        Sends a Gotify notification for git updates

        :param new_version: The new version available from git
        """
        if settings.USE_GOTIFY:
            update_text = notifyStrings[NOTIFY_UPDATE_TEXT]
            title = notifyStrings[NOTIFY_UPDATE]
            self._notify_gotify(title, update_text + new_version)

    def notify_login(self, ipaddress=""):
        """
        Sends a Gotify notification on login

        :param ipaddress: The ip address the login is originating from
        """
        if settings.USE_GOTIFY:
            update_text = notifyStrings[NOTIFY_LOGIN_TEXT]
            title = notifyStrings[NOTIFY_LOGIN]
            self._notify_gotify(title, update_text.format(ipaddress))

    def _notify_gotify(self, title, message, host=None, token=None, force=False):
        """
        Sends a Gotify notification

        :param title: The title of the notification to send
        :param message: The message string to send
        :param host: The host where Gotify is running (incl port)
        :param token: Gotify authorization token
        :param force: Enforce sending, for instance for testing

        :returns: the message to send
        """

        if not (force or settings.USE_GOTIFY):
            logger.debug("Notification for Gotify not enabled, skipping this notification")
            return False, "Disabled"

        logger.debug("Sending a Gotify message {0}".format(message))

        return self._send_gotify_msg(title, message, host, token)



"""
Test Provider Result Parsing
"""

import re
import sys
import unittest
from functools import wraps

import pytest
import validators

import sickchill.oldbeard.providers
from sickchill import movies, settings

pytest.skip(allow_module_level=True)

settings.CPU_PRESET = "NORMAL"


disabled_provider_tests = {
    # ???
    "Cpasbien": ["rss", "episode", "season"],
    "SkyTorrents": ["rss", "episode", "season", "cache", "result"],
    "ilCorsaroNero": ["rss"],
    # api_maintenance still
    "TorrentProject": ["rss", "episode", "season", "cache", "result"],
    # Have to trick it into thinking is an anime search, and add string overrides
    "TokyoToshokan": ["rss", "episode", "season"],
    "LimeTorrents": ["rss", "episode", "season"],
    "KickAssTorrents": ["rss", "episode", "season", "cache", "result"],
    "Torrentz": ["rss", "episode", "season", "cache", "result"],
    "ThePirateBay": ["rss", "episode", "season", "cache", "result"],
    "EZTV": ["rss", "episode", "season", "cache", "result"],
    "Rarbg": ["season", "episode", "rss", "movie"],
    # Demonoid is onion only now
    "Demonoid": ["rss", "episode", "season", "cache", "result"],
    # HorribleSubs needs rewritten
    "HorribleSubs": ["rss", "episode", "season", "cache", "result"],
    "Nyaa": ["rss"],
}

test_string_overrides = {
    "Cpasbien": {"Episode": ["The 100 S07E08"], "Season": ["The 100 S06"]},
    "Torrent9": {"Episode": ["Power Book IV Force S01E01"], "Season": ["The Lost Symbol S01"]},
    "Nyaa": {"Episode": ["Fairy Tail S2"], "Season": ["Fairy Tail S2"]},
    "TokyoToshokan": {"Episode": ["Fairy Tail S2"], "Season": ["Fairy Tail S2"]},
    "HorribleSubs": {"Episode": ["Fairy Tail S2"], "Season": ["Fairy Tail S2"]},
    "Demonoid": {"Episode": ["Star Trek Picard S01E04"], "Season": ["Locke and Key 2020 S01"]},
    "ilCorsaroNero": {"Episode": ["The 100 S06E02"]},
    "Rarbg": {"Season": ["Person of Interest S02"], "Episode": ["WandaVision S01E09"]},
}

magnet_regex = re.compile(r"magnet:\?xt=urn:btih:\w{32,40}(:?&dn=[\w. %+-]+)*(:?&tr=(:?tcp|https?|udp)[\w%. +-]+)*")


class BaseParser(type):
    class TestCase(unittest.TestCase):
        provider = None

        def __init__(self, test):
            """Initialize the test suite"""
            super().__init__(test)

            self.provider.session.verify = True
            settings.SSL_VERIFY = True

            self.provider.username = self.username
            self.provider.password = self.password
            settings.movie_list = movies.MovieList()

        @property
        def username(self):
            # TODO: Make this read usernames from somewhere
            return ""

        @property
        def password(self):
            # TODO: Make this read passwords from somewhere
            return ""

        def search_strings(self, mode):
            _search_strings = {"RSS": [""], "Episode": ["The 100 S07E08"], "Season": ["Game of Thrones S08"], "Movie": ["Black Panther 2018"]}
            _search_strings.update(self.provider.cache.search_params)
            _search_strings.update(test_string_overrides.get(self.provider.name, {}))
            return {mode: _search_strings[mode]}

        def magic_skip(func):  # pylint:disable=no-self-argument
            @wraps(func)
            def magic(self, *args, **kwargs):
                if func.__name__.split("_")[1] in disabled_provider_tests.get(self.provider.name, []):
                    self.skipTest("Test is programmatically disabled for provider {}".format(self.provider.name))
                func(self, *args, **kwargs)

            return magic

        def shortDescription(self):
            if self._testMethodDoc:
                return self._testMethodDoc.replace("the provider", self.provider.name)
            return None

        @magic_skip
        def test_rss_search(self):
            """Check that the provider parses rss search results"""
            if self.provider.can_daily:
                results = self.provider.search(self.search_strings("RSS"))
                assert results, results

        @magic_skip
        def test_episode_search(self):
            """Check that the provider parses episode search results"""
            if self.provider.can_backlog:
                results = self.provider.search(self.search_strings("Episode"))
                assert results, results

        @magic_skip
        def test_season_search(self):
            """Check that the provider parses season search results"""
            if self.provider.can_backlog:
                results = self.provider.search(self.search_strings("Season"))
                assert results, results

        # @pytest.mark.skip("Need to add a movie to the database before movie search works")
        @magic_skip
        def test_movie_search(self):
            """Check that the provider parses episode search results"""
            if self.provider.supports_movies:
                results = self.provider.search(self.search_strings("Movie"))
                assert results, results

        @magic_skip
        def test_cache_update(self):
            """Check that the provider's cache parses rss search results"""
            if self.provider.can_daily:
                self.provider.cache.update_cache()

        @magic_skip
        def test_result_values(self):
            """Check that the provider returns results in proper format"""
            if self.provider.can_backlog:
                for result in self.provider.search(self.search_strings("Episode")):
                    self.assertIsInstance(result, dict)
                    assert sorted(result) == ["hash", "leechers", "link", "seeders", "size", "title"]

                    self.assertIsInstance(result["title"], str)
                    self.assertIsInstance(result["link"], str)
                    self.assertIsInstance(result["hash"], str)
                    self.assertIsInstance(result["seeders"], int)
                    self.assertIsInstance(result["leechers"], int)
                    self.assertIsInstance(result["size"], int)

                    assert len(result["title"])
                    assert len(result["link"])
                    assert len(result["hash"]) in (0, 32, 40)
                    assert result["seeders"] >= 0
                    assert result["leechers"] >= 0

                    assert result["size"] >= -1

                    if result["link"].startswith("magnet"):
                        assert magnet_regex.match(result["link"])
                    else:
                        assert self.provider.valid_url(result["link"]), result["link"]

                    self.assertIsInstance(self.provider._get_size(result), int)
                    assert all(self.provider._get_title_and_url(result))
                    assert self.provider._get_size(result)

        @pytest.mark.skip("Not yet implemented")
        def test_season_search_strings_format(self):
            """Check format of the provider's season search strings"""
            pass

        @pytest.mark.skip("Not yet implemented")
        def test_episode_search_strings_format(self):
            """Check format of the provider's season search strings"""
            pass


def generate_test_cases():
    """
    Auto Generate TestCases from providers and add them to globals()
    """
    for p in sickchill.oldbeard.providers.__all__:
        provider = sickchill.oldbeard.providers.getProviderModule(p).Provider()
        if provider.can_backlog and provider.provider_type == "torrent" and provider.public:
            generated_class = type(str(provider.name), (BaseParser.TestCase,), {"provider": provider})
            globals()[generated_class.__name__] = generated_class
            del generated_class


generate_test_cases()

if __name__ == "__main__":
    import inspect

    print("=====> Testing %s", __file__)

    def override_log(msg, *args, **kwargs):
        """Override the SickChill logger so we can see the debug output from providers"""
        _ = args, kwargs
        print(msg)

    sickchill.logger.info = override_log
    sickchill.logger.debug = override_log
    sickchill.logger.error = override_log
    sickchill.logger.warning = override_log

    suite = unittest.TestSuite()
    members = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    for _, provider_test_class in members:
        if provider_test_class not in (BaseParser, BaseParser.TestCase):
            suite.addTest(unittest.TestLoader().loadTestsFromTestCase(provider_test_class))

    unittest.TextTestRunner(verbosity=3).run(suite)



import datetime
import operator
import re
import threading
import time
import traceback

from sickchill import logger, oldbeard, settings
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard.scdatetime import sctimeago
from sickchill.show.History import History

from . import db, helpers
from .common import cpu_presets, DOWNLOADED, Quality, SNATCHED, SNATCHED_PROPER
from .name_parser.parser import InvalidNameException, InvalidShowException, NameParser
from .search import pick_best_result, snatch_episode


class ProperFinder(object):
    def __init__(self):
        self.amActive = False

    # noinspection PyUnusedLocal
    def run(self, force=False):
        """
        Start looking for new propers

        :param force: Start even if already running (currently not used, defaults to False)
        """
        logger.info(_("Beginning the search for new propers"))

        self.amActive = True

        propers = self._get_propers()

        if propers:
            self._download_propers(propers)

        self._set_last_proper_search(datetime.datetime.today().toordinal())

        when = sctimeago(-settings.properFinderScheduler.timeLeft())
        logger.info(_("Completed the search for new propers, next check approximately {when}").format(when=when))

        self.amActive = False

    def _get_propers(self):
        """
        Walk providers for propers
        """
        propers = {}

        search_date = datetime.datetime.today() - datetime.timedelta(days=2)

        # for each provider get a list of the proper releases
        original_thread_name = threading.current_thread().name
        providers = [x for x in oldbeard.providers.sorted_provider_list(settings.RANDOMIZE_PROVIDERS) if x.is_active]
        for provider in providers:
            threading.current_thread().name = f"{original_thread_name} :: [{provider.name}]"

            logger.info(_("Searching for any new PROPER releases from {name}").format(name=provider.name))

            try:
                provider_propers = provider.find_propers(search_date)
            except AuthException as error:
                logger.warning(_("Authentication error: {error}").format(error=error))
                continue
            except Exception as error:
                logger.exception(_("Exception while searching propers in {name}, skipping: {error}").format(name=provider.name, error=error))
                logger.debug(traceback.format_exc())
                continue

            # if they haven't been added by a different provider than add the proper to the list
            for proper in provider_propers:
                if not re.search(r"\b(proper|repack|real)\b", proper.name, re.I):
                    logger.debug(_("find_propers returned a non-proper, we have caught and skipped it."))
                    continue

                name = self._generic_name(proper.name)
                if name not in propers:
                    logger.debug(_("Found new proper: {name}").format(name=proper.name))
                    proper.provider = provider
                    propers[name] = proper

            threading.current_thread().name = original_thread_name

        # take the list of unique propers and get it sorted by date
        sorted_propers = sorted(list(propers.values()), key=operator.attrgetter("date"), reverse=True)
        final_propers = []

        for proper in sorted_propers:
            try:
                parse_result = NameParser(False).parse(proper.name)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug(_("Error parsing {name}: {error}").format(name=proper.name, error=error))
                continue

            if not parse_result.series_name:
                continue

            if parse_result.show.paused:
                logger.debug(_("Ignoring {name} because {show} is paused").format(name=proper.name, show=proper.show.name))
                continue

            if not parse_result.episode_numbers:
                logger.debug(_("Ignoring {name} because it's for a full season rather than specific episode").format(name=proper.name))
                continue

            logger.debug(
                _("Successful match! Result {name} matched to show {show_name}").format(name=parse_result.original_name, show_name=parse_result.show.name)
            )

            # set the indexerid in the db to the show's indexerid
            proper.indexerid = parse_result.show.indexerid

            # set the indexer in the db to the show's indexer
            proper.indexer = parse_result.show.indexer

            # populate our Proper instance
            proper.show = parse_result.show
            proper.season = parse_result.season_number if parse_result.season_number is not None else 1
            proper.episode = parse_result.episode_numbers[0]
            proper.release_group = parse_result.release_group
            proper.version = parse_result.version
            proper.quality = Quality.nameQuality(proper.name, parse_result.is_anime)
            proper.content = None

            # filter release
            best_result = pick_best_result(proper, parse_result.show)
            if not best_result:
                logger.debug(_("Proper {name} were rejected by our release filters").format(name=proper.name))
                continue

            # only get anime proper if it has release group and version
            if best_result.show.is_anime and not best_result.release_group and best_result.version == -1:
                logger.debug(_("Proper {name} doesn't have a release group and version, ignoring it").format(name=best_result.name))
                continue

            # check if we actually want this proper (if it's the right quality)
            main_db_con = db.DBConnection()
            sql_results = main_db_con.select(
                "SELECT status FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?",
                [best_result.indexerid, best_result.season, best_result.episode],
            )
            if not sql_results:
                continue

            # only keep the proper if we have already retrieved the same quality ep (don't get better/worse ones)
            old_status, old_quality = Quality.splitCompositeStatus(int(sql_results[0]["status"]))
            if old_status not in (DOWNLOADED, SNATCHED) or old_quality != best_result.quality:
                continue

            # check if we actually want this proper (if it's the right release group and a higher version)
            if best_result.show.is_anime:
                main_db_con = db.DBConnection()
                sql_results = main_db_con.select(
                    "SELECT release_group, version FROM tv_episodes WHERE showid = ? AND season = ? AND episode = ?",
                    [best_result.indexerid, best_result.season, best_result.episode],
                )

                old_version = int(sql_results[0]["version"])
                old_release_group = sql_results[0]["release_group"]

                if -1 < old_version < best_result.version:
                    logger.info(_("Found new anime v{version} to replace existing v{old_version}").format(version=best_result.version, old_version=old_version))
                else:
                    continue

                if old_release_group != best_result.release_group:
                    logger.info(
                        "Skipping proper from release group: {release_group}, does not match existing release group: {old_release_group}".format(
                            release_group=best_result.release_group, old_release_group=old_release_group
                        )
                    )
                    continue

            # if the show is in our list and there hasn't been a proper already added for that particular episode then add it to our list of propers
            if best_result.indexerid != -1 and (best_result.indexerid, best_result.season, best_result.episode) not in {
                (p.indexerid, p.season, p.episode) for p in final_propers
            }:
                logger.info(_("Found a proper that we need: {name}").format(name=best_result.name))
                final_propers.append(best_result)

        return final_propers

    def _download_propers(self, proper_list):
        """
        Download proper (snatch it)

        :param proper_list:
        """

        for proper in proper_list:
            history_age = datetime.datetime.today() - datetime.timedelta(days=30)

            # make sure the episode has been downloaded before
            main_db_con = db.DBConnection()
            history_results = main_db_con.select(
                "SELECT resource FROM history "
                + "WHERE showid = ? AND season = ? AND episode = ? AND quality = ? AND date >= ? "
                + "AND action IN ("
                + ",".join([str(x) for x in Quality.SNATCHED + Quality.DOWNLOADED])
                + ")",
                [proper.indexerid, proper.season, proper.episode, proper.quality, history_age.strftime(History.date_format)],
            )

            # if we didn't download this episode in the first place we don't know what quality to use for the proper, so we can't do it
            if not history_results:
                logger.info(_("Unable to find an original history entry for proper {name} so I'm not downloading it.").format(name=proper.name))
                continue

            else:
                # make sure that none of the existing history downloads are the same proper we're trying to download
                clean_proper_name = self._generic_name(helpers.remove_non_release_groups(proper.name))
                is_same = False
                for result in history_results:
                    # if the result exists in history already we need to skip it
                    if self._generic_name(helpers.remove_non_release_groups(result["resource"])) == clean_proper_name:
                        is_same = True
                        break
                if is_same:
                    logger.debug(_("This proper is already in history, skipping it"))
                    continue

                # get the episode object
                episode_object = proper.show.get_episode(proper.season, proper.episode)

                # make the result object
                result = proper.provider.get_result([episode_object], proper.url)
                result.show = proper.show
                result.name = proper.name
                result.quality = proper.quality
                result.release_group = proper.release_group
                result.version = proper.version
                result.content = proper.content

                # snatch it
                snatch_episode(result, SNATCHED_PROPER)
                time.sleep(cpu_presets[settings.CPU_PRESET])

    @staticmethod
    def _generic_name(name):
        return name.replace(".", " ").replace("-", " ").replace("_", " ").lower()

    @staticmethod
    def _set_last_proper_search(when):
        """
        Record last proper search in DB

        :param when: When was the last proper search
        """

        logger.debug(_("Setting the last Proper search in the DB to {when}").format(when=when))

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT last_proper_search FROM info")

        if not sql_results:
            main_db_con.action("INSERT INTO info (last_backlog, last_indexer, last_proper_search) VALUES (?,?,?)", [0, 0, str(when)])
        else:
            main_db_con.action(f"UPDATE info SET last_proper_search = ? WHERE 1", [str(when)])

    @staticmethod
    def _get_last_proper_search():
        """
        Find last proper search from DB
        """

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT last_proper_search FROM info")

        last_proper_search = datetime.date.min

        # noinspection PyBroadException
        try:
            last_proper_search = datetime.date.fromordinal(int(sql_results[0]["last_proper_search"]))
        except Exception:
            pass

        return last_proper_search



import socket
import sys
import threading
import zlib
from time import sleep, time
from typing import Any, Dict

from .aniDBcommands import Command
from .aniDBerrors import AniDBBannedError, AniDBError, AniDBMustAuthError, AniDBPacketCorruptedError
from .aniDBresponses import ResponseResolver


class AniDBLink(threading.Thread):
    def __init__(self, server, port, myport, logFunction, delay=2, timeout=20, log_private=False):
        super().__init__()
        self.server = server
        self.port = port
        self.target = (server, port)
        self.timeout = timeout

        self.name = "ANIDB-LINK"

        self.myport = myport
        self.sock = None
        self.bound = self.connectSocket(myport, self.timeout)

        self.cmd_queue: Dict[Any, Command] = {None: None}
        self.resp_tagged_queue = {}
        self.resp_untagged_queue = []
        self.tags = []
        self.lastpacket = time()
        self.delay = delay
        self.session = None
        self.banned = False
        self.crypt = None

        self.log = logFunction
        self.log_private = log_private

        self._stop = threading.Event()
        self._quiting = False
        self.daemon = True
        self.start()

    def connectSocket(self, myport, timeout):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(timeout)
        portlist = [myport] + [7654]
        for port in portlist:
            try:
                self.sock.bind(("", port))
            except Exception:
                continue
            else:
                self.myport = port
                return True
        else:
            return False

    def disconnectSocket(self):
        if self.sock:
            self.sock.close()

    def stop(self):
        self.log("Releasing socket and stopping link thread")
        self._quiting = True
        self.disconnectSocket()
        self._stop.set()

    def stopped(self):
        return self._stop.is_set()

    @staticmethod
    def print_log(data):
        print(data)

    def print_log_dummy(self, data):
        pass

    def run(self):
        while not self._quiting:
            try:
                data = self.sock.recv(8192)
            except socket.timeout:
                self._handle_timeouts()
                continue

            self.log("NetIO < %s" % repr(data))
            try:
                for i in range(2):
                    try:
                        tmp = data
                        resp = None
                        if tmp[:2] == b"\x00\x00":
                            self.log("Attempting inflation")
                            tmp = zlib.decompressobj().decompress(tmp[2:])
                            self.log("UnZip | %s" % repr(tmp))
                        self.log("Decoding")
                        tmp = tmp.decode("utf8")
                        self.log("Parsing")
                        resp = ResponseResolver(tmp)
                        self.log("Response success!")
                    except Exception:
                        self.log("ResponseResolver Error")
                        sys.excepthook(*sys.exc_info())
                        self.crypt = None
                        self.session = None
                    else:
                        break
                if not resp:
                    raise AniDBPacketCorruptedError("Either decrypting, decompressing or parsing the packet failed")
                cmd = self._cmd_dequeue(resp)
                resp = resp.resolve(cmd)
                resp.parse()
                if resp.rescode in ("200", "201"):
                    self.session = resp.attrs["sesskey"]
                if resp.rescode in ("209",):
                    print("sorry encryption is not supported")
                    raise Exception
                    # self.crypt=aes(md5(resp.req.apipassword+resp.attrs['salt']).digest())
                if resp.rescode in ("203", "403", "500", "501", "503", "506"):
                    self.session = None
                    self.crypt = None
                if resp.rescode in ("504", "555"):
                    self.banned = True
                    print("AniDB API informs that user or client is banned:", resp.resstr)
                resp.handle()
                if not cmd or not cmd.mode:
                    self._resp_queue(resp)
                else:
                    self.tags.remove(resp.restag)
            except Exception:
                self.log("Catastrophic error - Closing anidb thread")
                sys.excepthook(*sys.exc_info())
                print("Avoiding flood by paranoidly panicing: Aborting link thread, killing connection, releasing waiters and quiting")
                self.disconnectSocket()
                try:
                    cmd.waiter.release()
                except Exception:
                    pass

                for tag, cmd in self.cmd_queue.items():
                    try:
                        cmd.waiter.release()
                    except Exception:
                        pass

                sys.exit()

    def _handle_timeouts(self):
        willpop = []
        for tag, cmd in self.cmd_queue.items():
            if not tag:
                continue
            if time() - cmd.started > self.timeout:
                self.tags.remove(cmd.tag)
                willpop.append(cmd.tag)
                cmd.waiter.release()

        for tag in willpop:
            self.cmd_queue.pop(tag)

    def _resp_queue(self, response):
        if response.restag:
            self.resp_tagged_queue[response.restag] = response
        else:
            self.resp_untagged_queue.append(response)

    def getresponse(self, command):
        if command:
            resp = self.resp_tagged_queue.pop(command.tag)
        else:
            resp = self.resp_untagged_queue.pop()
        self.tags.remove(resp.restag)
        return resp

    def _cmd_queue(self, command):
        self.cmd_queue[command.tag] = command
        self.tags.append(command.tag)

    def _cmd_dequeue(self, resp) -> Command:
        if not resp.restag:
            return None
        else:
            return self.cmd_queue.pop(resp.restag)

    def _delay(self):
        return self.delay < 2.1 and 2.1 or self.delay

    def _do_delay(self):
        age = time() - self.lastpacket
        delay = self._delay()
        if age <= delay:
            sleep(delay - age)

    def _send(self, command):
        if self.banned:
            self.log("NetIO | BANNED")
            raise AniDBBannedError("Not sending, banned")
        self._do_delay()
        self.lastpacket = time()
        command.started = time()
        data = command.raw_data()

        encoded_data = data.encode("utf-8")

        self.sock.sendto(encoded_data, self.target)

        if command.command == "AUTH" and self.log_private:
            self.log("NetIO > sensitive data is not logged!")
        else:
            self.log("NetIO > %s" % repr(data))

    def new_tag(self):
        if not len(self.tags):
            maxtag = "T000"
        else:
            maxtag = max(self.tags)
        newtag = "T%03d" % (int(maxtag[1:]) + 1)
        return newtag

    def request(self, command):
        if not self.sock and self.connectSocket(self.myport, self.timeout):
            self.log("Not connected to aniDB, not sending command")
            raise AniDBError("No Socket")

        if not (self.session and command.session) and command.command not in ("AUTH", "PING", "ENCRYPT"):
            raise AniDBMustAuthError("You must be authed to execute commands besides AUTH and PING")
        command.started = time()
        self._cmd_queue(command)
        self._send(command)



import warnings

from bs4 import BeautifulSoup, FeatureNotFound


class ParserBeautifulSoup(BeautifulSoup):
    """A ``bs4.BeautifulSoup`` that picks the first parser available in `parsers`.

    :param markup: markup for the ``bs4.BeautifulSoup``.
    :param list parsers: parser names, in order of preference.

    Snagged from subliminal, because of course it was.
    """

    def __init__(self, markup, parsers, **kwargs):
        # reject features
        if set(parsers).intersection({"fast", "permissive", "strict", "xml", "html", "html5"}):
            raise ValueError("Features not allowed, only parser names")

        # reject some kwargs
        if "features" in kwargs:
            raise ValueError("Cannot use features kwarg")
        if "builder" in kwargs:
            raise ValueError("Cannot use builder kwarg")

        # pick the first parser available
        for parser in parsers:
            try:
                super(ParserBeautifulSoup, self).__init__(markup, parser, **kwargs)
                return
            except FeatureNotFound:
                pass

        raise FeatureNotFound


class BS4Parser(object):
    def __init__(self, markup, parsers=None, language="html", **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(parsers, str):
                parsers = [parsers]

            if language == "html":
                extra_parsers = ["lxml", "html5lib", "html.parser"]
            elif language == "xml":
                extra_parsers = ["lxml-xml", "html.parser"]

            if parsers is None:
                parsers = extra_parsers

            try:
                self.soup = ParserBeautifulSoup(markup, parsers, **kwargs)
            except FeatureNotFound:
                self.soup = ParserBeautifulSoup(markup, extra_parsers, **kwargs)

    def __enter__(self) -> ParserBeautifulSoup:
        return self.soup

    def __exit__(self, exc_ty, exc_val, tb):
        self.soup.clear(True)
        self.soup = None



"""
Test history
"""

import unittest


class HistoryTests(unittest.TestCase):
    """
    Test history
    """


if __name__ == "__main__":
    print("=====> Testing {0}".format(__file__))

    SUITE = unittest.TestLoader().loadTestsFromTestCase(HistoryTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import multiprocessing
import os.path
import sys

import servicemanager
import SickChill
import win32service
import win32serviceutil


class SickChillService(win32serviceutil.ServiceFramework):
    _svc_name_ = "sickchill"
    _svc_display_name_ = "sickchill service"
    _svc_description_ = "Runs sickchill web service in the background."
    _exe_name_ = sys.executable
    _exe_args_ = f"-u -E {os.path.abspath(__file__)}"

    proc = None

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        if self.proc:
            self.proc.terminate()

    def SvcRun(self):
        self.proc = multiprocessing.Process(target=SickChill.main)
        self.proc.start()
        self.ReportServiceStatus(win32service.SERVICE_RUNNING)
        self.SvcDoRun()
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

    def SvcDoRun(self):
        self.proc.join()


if __name__ == "__main__":
    # noinspection PyBroadException
    try:
        if len(sys.argv) == 1:
            import win32traceutil

            servicemanager.Initialize()
            servicemanager.PrepareToHostSingle(SickChillService)
            servicemanager.StartServiceCtrlDispatcher()
        elif "--fg" in sys.argv:
            sys.argv.remove("--fg")
            SickChill.main()
        else:
            win32serviceutil.HandleCommandLine(SickChillService)
    except (SystemExit, KeyboardInterrupt):
        raise
    except Exception:
        import traceback

        traceback.print_exc()



import copy
import re
from base64 import b16encode, b32decode
from datetime import datetime
from itertools import chain
from os.path import join
from random import shuffle
from typing import Callable, Dict, Iterable, List, TYPE_CHECKING, Union
from urllib.parse import urljoin

import validators
from requests.structures import CaseInsensitiveDict
from requests.utils import add_dict_to_cookiejar

import sickchill.oldbeard
from sickchill import logger
from sickchill.helper.common import sanitize_filename
from sickchill.oldbeard import filters
from sickchill.oldbeard.classes import Proper, SearchResult
from sickchill.oldbeard.common import MULTI_EP_RESULT, Quality, SEASON_RESULT
from sickchill.oldbeard.db import DBConnection
from sickchill.oldbeard.helpers import download_file, getURL, make_session, remove_file_failed
from sickchill.oldbeard.name_parser.parser import InvalidNameException, InvalidShowException, NameParser
from sickchill.oldbeard.show_name_helpers import allPossibleShowNames
from sickchill.oldbeard.tvcache import TVCache

if TYPE_CHECKING:
    from sickchill.tv import TVEpisode


class GenericProvider(object):
    NZB = "nzb"
    NZBDATA = "nzbdata"
    TORRENT = "torrent"

    PROVIDER_BROKEN = 0
    PROVIDER_DAILY = 1
    PROVIDER_BACKLOG = 2
    PROVIDER_OK = 3

    ProviderStatus = {
        PROVIDER_BROKEN: _("Not working"),
        PROVIDER_DAILY: _("Daily/RSS only"),
        PROVIDER_BACKLOG: _("Backlog/Manual Search only"),
        PROVIDER_OK: _("Daily/RSS and Backlog/Manual Searches working"),
    }

    def __init__(self, name):
        self.__original_url = None
        self.__original_urls = None

        self.name = name

        self.anime_only = False
        self.bt_cache_urls = [
            # "http://torcache.net/torrent/{torrent_hash}.torrent",
            # "http://torrentproject.se/torrent/{torrent_hash}.torrent",
            # "http://thetorrent.org/torrent/{torrent_hash}.torrent",
            # "http://btdig.com/torrent/{torrent_hash}.torrent",
            ("https://torrage.info/download?h={torrent_hash}", "https://torrage.info/torrent.php?h={torrent_hash}"),
            "https://itorrents.org/torrent/{torrent_hash}.torrent?title={torrent_name}",
        ]
        self.cache = TVCache(self)

        self.current_episode_object = None

        self.enable_backlog = False
        self.enable_daily = False
        self.enabled = False
        self.headers = CaseInsensitiveDict()
        self.proper_strings = ["PROPER|REPACK|REAL"]
        self.provider_type = None
        self.public = False
        self.search_fallback = False
        self.search_mode = None
        self.session = make_session()
        self.show = None
        self.supports_absolute_numbering = False
        self.supports_backlog = True
        self.supports_movies = False

        self.url = ""
        self.urls = {}

        # Use and configure the attribute enable_cookies to show or hide the cookies input field per provider
        self.enable_cookies = False
        self.cookies = ""
        self.rss_cookies = ""

        self.ability_status = self.PROVIDER_OK

        self.size_units = ["B", "KB", "MB", "GB", "TB", "PB"]

        shuffle(self.bt_cache_urls)

    def download_result(self, result):
        if not self.login():
            return False

        urls, filename = self._make_url(result)

        for url in urls:
            if "NO_DOWNLOAD_NAME" in url:
                continue

            if isinstance(url, tuple):
                referer = url[1]
                url = url[0]
            else:
                referer = "/".join(url.split("/")[:3]) + "/"

            if url.startswith("http"):
                self.headers.update({"Referer": referer})

            logger.info(f"Downloading a result from {self.name} at {url}")

            downloaded_filename = download_file(
                url, filename, session=self.session, headers=self.headers, hooks={"response": self.get_url_hook}, return_filename=True
            )
            if downloaded_filename:
                if self._verify_download(downloaded_filename):
                    logger.info(f"Saved result to {downloaded_filename}")
                    return True

                logger.warning(f"Could not download {url}")
                remove_file_failed(downloaded_filename)

        if urls:
            logger.warning(f"{self.name} : Failed to download any results")

        return False

    def find_propers(self, search_date=None):
        results = self.cache.list_propers(search_date)

        return [Proper(x["name"], x["url"], datetime.fromtimestamp(x["time"]), self.show) for x in results]

    def find_search_results(self, show, episodes, search_mode, manual_search=False, download_current_quality=False):
        self._check_auth()
        self.show = show

        results = {}
        items_list = []
        searched_scene_season = None

        for episode in episodes:
            cache_result = self.cache.search_cache(episode, manual_search=manual_search, down_cur_quality=download_current_quality)
            if cache_result:
                if episode.episode not in results:
                    results[episode.episode] = cache_result
                else:
                    results[episode.episode].extend(cache_result)

                continue

            if len(episodes) > 1 and search_mode == "season" and searched_scene_season == episode.scene_season:
                continue

            search_strings = []
            searched_scene_season = episode.scene_season

            if len(episodes) > 1 and search_mode == "season":
                search_strings = self.get_season_search_strings(episode)
            elif search_mode == "episode":
                search_strings = self.get_episode_search_strings(episode)

            self.current_episode_object = episode

            for search_string in search_strings:
                items_list += self.search(search_string)

        if len(results) == len(episodes):
            return results

        if items_list:
            items = {}
            unknown_items = []

            for item in items_list:
                quality = self.get_quality(item, anime=show.is_anime)

                if quality == Quality.UNKNOWN:
                    unknown_items.append(item)
                elif quality == Quality.NONE:
                    pass  # Skipping an HEVC when HEVC is not allowed by settings
                else:
                    if quality not in items:
                        items[quality] = []
                    items[quality].append(item)

            items_list = list(chain(*[v for (k_, v) in sorted(items.items(), reverse=True)]))
            items_list += unknown_items

        cache_list = []

        for item in items_list:
            title, url = self._get_title_and_url(item)
            seeders, leechers = self._get_seeders_and_leechers(item)
            size = self._get_size(item)

            try:
                parse_result = NameParser(parse_method=("normal", "anime")[show.is_anime]).parse(title)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug(f"{error}")
                continue

            show_object = parse_result.show
            quality = parse_result.quality
            release_group = parse_result.release_group
            version = parse_result.version
            skip_release = False
            actual_episodes = []
            actual_season = -1

            if not (show_object.air_by_date or show_object.sports):
                if search_mode == "season":
                    if parse_result.episode_numbers:
                        logger.debug(f"This is supposed to be a season pack search but the result {title} is not a valid season pack, skipping it")
                        skip_release = True
                    elif not [ep for ep in episodes if parse_result.season_number == (ep.season, ep.scene_season)[ep.show.is_scene]]:
                        logger.info(f"This season result {title} is for a season we are not searching for, skipping it")
                        skip_release = True

                else:
                    if not all(
                        [
                            parse_result.season_number is not None,
                            parse_result.episode_numbers,
                            [ep for ep in episodes if ep.season == parse_result.season_number and ep.episode in parse_result.episode_numbers],
                        ]
                    ) and not all(
                        [
                            # fallback for anime on absolute numbering
                            parse_result.is_anime,
                            parse_result.ab_episode_numbers is not None,
                            [ep for ep in episodes if ep.show.is_anime and ep.absolute_number in parse_result.ab_episode_numbers],
                        ]
                    ):
                        logger.info(f"The result {title} doesn't seem to match an episode that we are currently trying to snatch, skipping it")
                        skip_release = True

                if not skip_release:
                    actual_season = parse_result.season_number
                    actual_episodes = parse_result.episode_numbers
            else:
                same_day_special = False

                if not parse_result.is_air_by_date:
                    logger.debug(f"This is supposed to be a date search but the result {title} didn't parse as one, skipping it")
                    skip_release = True
                else:
                    air_date = parse_result.air_date.toordinal()
                    db = DBConnection()
                    sql_results = db.select("SELECT season, episode FROM tv_episodes WHERE showid = ? AND airdate = ?", [show_object.indexerid, air_date])

                    if len(sql_results) == 2:
                        if int(sql_results[0]["season"]) == 0 and int(sql_results[1]["season"]) != 0:
                            actual_season = int(sql_results[1]["season"])
                            actual_episodes = [int(sql_results[1]["episode"])]
                            same_day_special = True
                        elif int(sql_results[1]["season"]) == 0 and int(sql_results[0]["season"]) != 0:
                            actual_season = int(sql_results[0]["season"])
                            actual_episodes = [int(sql_results[0]["episode"])]
                            same_day_special = True
                    elif len(sql_results) != 1:
                        logger.warning("Tried to look up the date for the episode {0} but the database didn't give proper results, skipping it".format(title))
                        skip_release = True

                    if not skip_release and not same_day_special:
                        actual_season = int(sql_results[0]["season"])
                        actual_episodes = [int(sql_results[0]["episode"])]

            logger.debug(f"Adding item from search to cache: {title}")

            cache_item = self.cache.add_cache_entry(title, url, size, seeders, leechers, parse_result=parse_result)

            if cache_item is not None:
                cache_list.append(cache_item)

            if skip_release:
                continue

            for episode_number in actual_episodes:
                if not show_object.want_episode(actual_season, episode_number, quality, manual_search, download_current_quality):
                    skip_release = True
                    break

            if skip_release:
                logger.debug(_("Ignoring result {title}.").format(title=title))
                continue

            logger.debug(_("Found result {title} at {url}.").format(title=title, url=url))

            episode_objects = []
            for current_episode in actual_episodes:
                episode_objects.append(show_object.get_episode(actual_season, current_episode))

            result = self.get_result(episode_objects, url)
            result.show = show_object
            result.name = title
            result.quality = quality
            result.release_group = release_group
            result.version = version
            result.content = None
            result.size = self._get_size(item)

            if len(episode_objects) == 1:
                episode_number = episode_objects[0].episode
                logger.debug("Single episode result.")
            elif len(episode_objects) > 1:
                episode_number = MULTI_EP_RESULT
                logger.debug(f"Separating multi-episode result to check for later - result contains episodes: {parse_result.episode_numbers}")
            elif len(episode_objects) == 0:
                episode_number = SEASON_RESULT
                logger.debug("Separating full season result to check for later")
            else:
                episode_number = None

            if episode_number not in results:
                results[episode_number] = [result]
            else:
                results[episode_number].append(result)

        if cache_list:
            # Access to a protected member of a client class
            cache_db = self.cache.get_db()
            cache_db.mass_upsert("results", cache_list)

        return results

    def get_id(self, suffix=""):
        return GenericProvider.make_id(self.name) + str(suffix)

    def get_quality(self, item, anime=False):
        (title, url_) = self._get_title_and_url(item)
        quality = Quality.scene_quality(title, anime)

        return quality

    def get_result(self, episodes, url):
        return self._get_result(episodes, self, url)

    # noinspection PyUnusedLocal
    @staticmethod
    def get_url_hook(response, **kwargs_):
        if response:
            logger.debug(f"{response.request.method} URL: {response.request.url} [Status: {response.status_code}]")

            if response.request.method == "POST":
                logger.debug(f"With post data: {response.request.body}")

    def get_url(self, url, post_data=None, params=None, timeout=30, **kwargs):
        kwargs["hooks"] = {"response": self.get_url_hook}
        return getURL(url, post_data=post_data, params=params, headers=self.headers, timeout=timeout, session=self.session, **kwargs)

    def image_name(self):
        return self.get_id() + ".png"

    @property
    def is_active(self):
        return False

    @property
    def is_enabled(self):
        return bool(self.enabled)

    @property
    def daily_enabled(self):
        return int(self.enable_daily)

    @property
    def backlog_enabled(self):
        return int(self.enable_backlog)

    @property
    def search_fallback_enabled(self):
        return int(self.search_fallback)

    @property
    def can_daily(self):
        return self.ability_status & self.PROVIDER_DAILY != 0

    @property
    def can_backlog(self):
        return self.ability_status & self.PROVIDER_BACKLOG != 0 and self.supports_backlog

    def status(self):
        return self.ProviderStatus.get(self.ability_status)

    @staticmethod
    def make_id(name):
        if not name:
            return ""

        return re.sub(r"[^\w_]", "_", str(name).strip().lower())

    def search_rss(self, episodes):
        return self.cache.find_needed_episodes(episodes)

    def seed_ratio(self):
        return ""

    def _check_auth(self):
        return True

    def login(self):
        return True

    def search(self, strings):
        return []

    def _get_result(self, episodes, provider, url):
        return SearchResult(episodes, provider, url)

    def get_episode_search_strings(self, episode: "TVEpisode", add_string: str = "") -> Union[List[Dict], Iterable]:
        if not episode:
            return []

        search_string = {"Episode": set()}

        for show_name in allPossibleShowNames(episode.show, season=episode.scene_season):
            episode_string = show_name + " "
            episode_string_fallback = None

            if episode.show.air_by_date:
                episode_string += str(episode.airdate).replace("-", " ")
            elif episode.show.sports:
                episode_string += str(episode.airdate).replace("-", " ")
                episode_string += ("|", " ")[len(self.proper_strings) > 1]
                episode_string += episode.airdate.strftime("%b")
            elif episode.show.anime:
                episode_string_fallback = episode_string + "{0:02d}".format(int(episode.scene_absolute_number))
                episode_string += "{0:03d}".format(int(episode.scene_absolute_number))
            else:
                episode_string += sickchill.oldbeard.config.naming_ep_type[2] % {
                    "seasonnumber": episode.scene_season,
                    "episodenumber": episode.scene_episode,
                }

            if add_string:
                episode_string += " " + add_string
                if episode_string_fallback:
                    episode_string_fallback += " " + add_string

            search_string["Episode"].add(episode_string.strip())
            if episode_string_fallback:
                search_string["Episode"].add(episode_string_fallback.strip())

        return [search_string]

    def get_season_search_strings(self, episode: "TVEpisode") -> List[Dict]:
        search_string = {"Season": set()}

        for show_name in allPossibleShowNames(episode.show, season=episode.scene_season):
            season_string = show_name + " "

            if episode.show.air_by_date or episode.show.sports:
                season_string += str(episode.airdate).split("-")[0]
            elif episode.show.anime:
                # use string below if you really want to search on season with number
                # season_string += 'Season ' + '{0:d}'.format(int(episode.scene_season))
                season_string += "Season"  # ignore season number to get all seasons in all formats
            else:
                season_string += "S{0:02d}".format(int(episode.scene_season))

            search_string["Season"].add(season_string.strip())

        return [search_string]

    def _get_size(self, item):
        try:
            return item.get("size", -1)
        except AttributeError:
            return -1

    def _get_seeders_and_leechers(self, item):
        try:
            return item.get("seeders", -1), item.get("leechers", -1)
        except AttributeError:
            return -1, -1

    def _get_storage_dir(self):
        return ""

    def _get_title_and_url(self, item):
        if not item:
            return "", ""

        title = item.get("title", "")
        url = item.get("link", "")

        if title:
            title = "" + title.replace(" ", ".")
        else:
            title = ""

        if url:
            url = url.replace("&amp;", "&").replace("%26tr%3D", "&tr=")
        else:
            url = ""

        return title, url

    @staticmethod
    def hash_from_magnet(magnet):
        # noinspection PyBroadException
        try:
            torrent_hash = re.findall(r"urn:btih:(\w{32,40})", magnet)[0].upper()
            if len(torrent_hash) == 32:
                torrent_hash = b16encode(b32decode(torrent_hash)).upper()
            return torrent_hash
        except Exception:
            logger.exception(f"Unable to extract torrent hash or name from magnet: {magnet}")
            return ""

    def _make_url(self, result):
        if not result:
            return "", ""

        filename = ""

        result.url = result.url.replace("http://itorrents.org", "https://itorrents.org")

        urls = [result.url]
        if result.url.startswith("magnet"):
            torrent_hash = self.hash_from_magnet(result.url)
            if not torrent_hash:
                return urls, filename

            # noinspection PyBroadException
            try:
                torrent_name = re.findall("dn=([^&]+)", result.url)[0]
            except Exception:
                torrent_name = "NO_DOWNLOAD_NAME"

            urls = []
            for cache_url in self.bt_cache_urls:
                if isinstance(cache_url, tuple):
                    urls.append(
                        (
                            cache_url[0].format(torrent_hash=torrent_hash, torrent_name=torrent_name),
                            cache_url[1].format(torrent_hash=torrent_hash, torrent_name=torrent_name),
                        )
                    )
                else:
                    urls.append(cache_url.format(torrent_hash=torrent_hash, torrent_name=torrent_name))

        if "torrage.info/torrent.php" in result.url:
            torrent_hash = result.url.split("=")[1]
            urls = [
                (
                    "https://t.torrage.info/download?h={torrent_hash}".format(torrent_hash=torrent_hash),
                    "https://torrage.info/torrent.php?h={torrent_hash}".format(torrent_hash=torrent_hash),
                )
            ]

        filename = join(self._get_storage_dir(), sanitize_filename(result.name) + "." + self.provider_type)

        return urls, filename

    def _verify_download(self, filename):
        return True

    def add_cookies_from_ui(self):
        """
        Adds the cookies configured from UI to the providers requests session
        :return: A tuple with the result (success result, and a descriptive message in str)
        """

        # This is the generic attribute used to manually add cookies for provider authentication
        if self.enable_cookies and self.cookies:
            cookie_validator = re.compile(r"^(\w+=[\w%-]+)(;\w+=[\w%-]+)*$")
            if not cookie_validator.match(self.cookies):
                return False, f"Cookie is not correctly formatted: {self.cookies}"

            new_cookies = {}
            for cookie in self.cookies.split(";"):
                key, value = cookie.rsplit("=", 1)
                new_cookies[key] = value

            add_dict_to_cookiejar(self.session.cookies, new_cookies)
            return True, "torrent cookie"

        return False, f"No Cookies added from ui for provider: {self.name}"

    def check_set_option(self, view, option, default="", cast: Callable = str, unhide=False):
        if hasattr(self, option):
            if view.request.method == "GET":
                value = view.get_query_argument(self.get_id("_" + option), default=default)
            elif view.request.method == "POST":
                value = view.get_body_argument(self.get_id("_" + option), default=default)
            else:
                raise AttributeError("We currently only work with GET and POST requests")

            if unhide:
                value = filters.unhide(getattr(self, option), value)

            return setattr(self, option, cast(value))

    @staticmethod
    def valid_url(url):
        try:
            valid = validators.url(url)
        except validators.utils.ValidationError:
            valid = False

        return valid is True

    @classmethod
    def invalid_url(cls, url):
        return not cls.valid_url(url)

    def __store_original_urls(self):
        if not self.__original_url:
            # make sure we keep a clean copy of the original base url, in case they reset the custom url
            self.__original_url = self.url

        if not self.__original_urls:
            # make sure we keep a clean copy of the original urls, in case they reset the custom url
            self.__original_urls = copy.deepcopy(self.urls)

    def __restore_original_urls(self):
        # custom_url removed/reset, set urls back to defaults
        if hasattr(self, "__original_urls"):
            self.urls = copy.deepcopy(self.__original_urls)
        if hasattr(self, "__original_url"):
            self.url = self.__original_url

    def check_and_update_urls(self):
        has_custom_url = False
        custom_url_valid = True
        if hasattr(self, "custom_url") and hasattr(self, "url") and hasattr(self, "urls"):
            has_custom_url = bool(self.custom_url)
            custom_url_valid = self.valid_url(self.custom_url)
            if has_custom_url and custom_url_valid:
                self.__store_original_urls()

                for url in self.__original_urls:
                    # update each url in our urls to have the correct proxy url
                    self.urls[url] = urljoin(self.custom_url, self.__original_urls[url].split(self.__original_url)[1])
            else:
                self.__restore_original_urls()

        return (True, custom_url_valid)[has_custom_url]



import datetime
import logging
from typing import List

import guessit
from slugify import slugify
from sqlalchemy import ForeignKey, JSON
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship, sessionmaker

logger = logging.getLogger("sickchill.movie")


class Base(DeclarativeBase):
    """Declarative Base Class"""


Session = sessionmaker()


class Movie(Base):
    __tablename__ = "movie"
    pk: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    date: Mapped[datetime.date]
    year: Mapped[int]
    status: Mapped[int]
    paused: Mapped[bool] = mapped_column(default=False)
    location: Mapped[str]
    start: Mapped[datetime.timedelta] = mapped_column(default=datetime.timedelta(days=-7))
    interval: Mapped[datetime.timedelta] = mapped_column(default=datetime.timedelta(days=1))
    added: Mapped[datetime.datetime] = mapped_column(default=datetime.datetime.now)
    updated: Mapped[datetime.datetime] = mapped_column(onupdate=datetime.datetime.now)
    completed: Mapped[datetime.datetime]
    searched: Mapped[datetime.datetime]
    slug: Mapped[str]

    language: Mapped[str]

    result_pk: Mapped[int] = mapped_column(ForeignKey("result.pk"))
    results: Mapped[List["Result"]] = relationship(backref="movie")

    images: Mapped[List["Images"]] = relationship(backref="movie")
    indexer_data: Mapped[List["IndexerData"]] = relationship(backref="movie")

    def __init__(self, name: str, year: int):
        self.name = name
        self.year = year

    @property
    def poster(self):
        return ""

    def __get_named_indexer_data(self, name):
        if self.indexer_data:
            for data in self.indexer_data:
                if data.site == name:
                    return data

    @property
    def imdb_data(self):
        data = self.__get_named_indexer_data("imdb")
        if data:
            return data.data
        return dict()

    @property
    def imdb_id(self):
        data = self.__get_named_indexer_data("imdb")
        if data:
            return data.pk
        return ""

    @property
    def tmdb_id(self):
        data = self.__get_named_indexer_data("tmdb")
        if data:
            return data.pk
        return ""

    @property
    def imdb_genres(self):
        data = self.__get_named_indexer_data("imdb")
        if data:
            return data.genres
        return []

    def __get_indexer_values(self, name, keys: list):
        try:
            data = getattr(self, f"{name}_data")
            for key in keys:
                data = data[key]
            return data
        except AttributeError:
            logger.debug(f"We do not have data for {name}")
        except (IndexError, KeyError):
            logger.debug(f"KeyError: {name}{''.join([f'[{k}]' for k in keys])}")

    @property
    def runtime(self):
        return self.__get_indexer_values("imdb", ["base", "runningTimeInMinutes"])

    @property
    def imdb_votes(self):
        return self.__get_indexer_values("imdb", ["ratings", "ratingCount"])

    @property
    def imdb_rating(self):
        return self.__get_indexer_values("imdb", ["ratings", "rating"])

    @property
    def imdb_outline(self):
        return self.__get_indexer_values("imdb", ["plot", "outline", "text"])

    @property
    def imdb_summary(self):
        return self.__get_indexer_values("imdb", ["plot", "summaries", 0, "text"])

    @staticmethod
    def slugify(target, value, old_value, initiator):
        if value and (not target.slug or value != old_value):
            target.slug = slugify(value)

    def search_strings(self):
        return {"Movie": [f"{self.name} {self.year}"]}

    def __repr__(self):
        return f"{self.name}"


listen(Movie.name, "set", Movie.slugify, retval=False)


class Result(Base):
    __tablename__ = "result"
    pk: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    title: Mapped[str]
    url: Mapped[str]
    size: Mapped[int]
    year: Mapped[int]
    provider: Mapped[str]
    seeders: Mapped[int]
    leechers: Mapped[int]
    info_hash: Mapped[str]
    group: Mapped[str]
    kind: Mapped[str]
    guess = mapped_column(JSON)
    found: Mapped[datetime.datetime] = mapped_column(default=datetime.datetime.now)
    updated: Mapped[datetime.datetime] = mapped_column(onupdate=datetime.datetime.now)

    session = Session()

    def __init__(self, result: dict, movie: Movie, provider):
        name = result["title"]
        guess = guessit.guessit(name)
        if not (guess and guess["type"] == "movie"):
            logging.debug(f"This is an episode, not a movie: {name}")
            return

        if not self.session.query(Movie).filter(Movie.name.like(f"{guess['title']}%")).count():
            logging.debug(f"This result does not match any of our movies")
            return

        self.info_hash = result["hash"]
        self.url = result["link"]
        self.name = name
        self.title = guess["title"]
        self.group = guess["release_group"]
        self.seeders = result["seeders"]
        self.leechers = result["leechers"]
        self.size = result["size"]
        self.year = guess["year"] or movie.year
        self.kind = provider.provider_type

        self.provider = provider.get_id()

        self.guess = guess

        self.movie = movie

    def __repr__(self):
        return f"{self.name}"


class Images(Base):
    __tablename__ = "images"

    url: Mapped[str] = mapped_column(primary_key=True)
    path: Mapped[str]
    site: Mapped[str]
    style: Mapped[int]

    movie_pk: Mapped[int] = mapped_column(ForeignKey("movie.pk"))

    def __init__(self, site: str, movie_pk: int, url: str, path: str, style: int):
        self.url = url
        self.path = path
        self.site = site
        self.style = style
        self.movie_pk = movie_pk


class IndexerData(Base):
    __tablename__ = "indexer_data"
    pk: Mapped[str] = mapped_column(primary_key=True)
    site: Mapped[str]
    data = mapped_column(JSON)

    movie_pk: Mapped[int] = mapped_column(ForeignKey("movie.pk"))

    genres: Mapped[List["Genres"]] = relationship(backref="indexer_data")

    def __repr__(self):
        return f"[{self.__tablename__.replace('_', ' ').title()}] {self.site}: {self.pk} - {self.movie.name}"


class Genres(Base):
    __tablename__ = "genres"
    pk: Mapped[str] = mapped_column(primary_key=True)
    indexer_data_pk: Mapped[int] = mapped_column(ForeignKey("indexer_data.pk"))



import json

import requests
from requests.structures import CaseInsensitiveDict

from sickchill import logger, settings
from sickchill.oldbeard import common


class Notifier(object):
    MATTERMOSTBOT_POST_URI = "/api/v4/posts"
    MATTERMOSTBOT_ICON_URL = "https://github.com/SickChill/SickChill/raw/master/sickchill/gui/slick/images/sickchill-sc.png"

    def notify_snatch(self, ep_name):
        if settings.MATTERMOSTBOT_NOTIFY_SNATCH:
            self._notify_mattermostbot(common.notifyStrings[common.NOTIFY_SNATCH], ep_name, "s")

    def notify_download(self, ep_name):
        if settings.MATTERMOSTBOT_NOTIFY_DOWNLOAD:
            self._notify_mattermostbot(common.notifyStrings[common.NOTIFY_DOWNLOAD], ep_name, "d")

    def notify_subtitle_download(self, ep_name, lang):
        if settings.MATTERMOSTBOT_NOTIFY_SUBTITLEDOWNLOAD:
            self._notify_mattermostbot(common.notifyStrings[common.NOTIFY_SUBTITLE_DOWNLOAD], ep_name + " - " + lang, "sd")

    def notify_update(self, new_version="??"):
        if settings.USE_MATTERMOSTBOT:
            update_text = common.notifyStrings[common.NOTIFY_UPDATE_TEXT]
            title = common.notifyStrings[common.NOTIFY_UPDATE]
            self._notify_mattermostbot("Update", title + " - " + update_text + new_version, "u")

    def notify_login(self, ipaddress=""):
        if settings.USE_MATTERMOSTBOT:
            update_text = common.notifyStrings[common.NOTIFY_LOGIN_TEXT]
            title = common.notifyStrings[common.NOTIFY_LOGIN]
            self._notify_mattermostbot("Login", title + " - " + update_text.format(ipaddress), "l")

    def test_notify(self):
        action = "Testing"
        message = "Serie name - S0E0 - Episode Name - Quality specs"
        tmsg = "t"
        return self._notify_mattermostbot(action, message, tmsg, force=True)

    def _send_mattermostbot(self, action="", message="", tmsg="t"):
        logger.info("Action: " + action + " | Message: " + message + " | tmsg: " + tmsg)
        mattermostbot_post_url = settings.MATTERMOSTBOT_URL + self.MATTERMOSTBOT_POST_URI
        mattermostbot_author = settings.MATTERMOSTBOT_AUTHOR
        mattermostbot_post_url.replace("//api", "/api")
        mattermostbot_icon_emoji = settings.MATTERMOSTBOT_ICON_EMOJI
        logger.info("Sending mattermost bot message: " + message + " towards " + mattermostbot_post_url)
        logger.info("Sending mattermost bot action: " + action)

        headers = CaseInsensitiveDict({"Content-Type": "application/json", "Authorization": "Bearer " + settings.MATTERMOSTBOT_TOKEN})
        contents = message.split(" - ")

        if tmsg == "s":
            color = "#ffff50"
            pretext = "Snatch/Start"
        elif tmsg == "d":
            color = "#009900"
            pretext = "Finish"
        elif tmsg == "t":
            color = "#0000ff"
            pretext = "Testing"
        else:
            color = "#ff0000"
            pretext = "Type: " + tmsg

        payload = {"channel_id": settings.MATTERMOSTBOT_CHANNEL, "message": "SickChill Work", "props": {"attachments": []}}
        payload["channel_id"] = settings.MATTERMOSTBOT_CHANNEL
        payload["props"]["attachments"] = []
        payload["props"]["attachments"].append({"pretext": pretext, "author_name": mattermostbot_author, "color": color, "fields": {}})
        payload["props"]["attachments"][0]["fields"] = []
        payload["props"]["attachments"][0]["fields"].append({"title": "Series", "value": contents[0], "short": "true"})
        payload["props"]["attachments"][0]["fields"].append({"title": "EP", "value": contents[1], "short": "true"})
        payload["props"]["attachments"][0]["fields"].append({"title": "Title", "value": contents[2], "short": "true"})
        payload["props"]["attachments"][0]["fields"].append({"title": "Quality", "value": contents[3], "short": "true"})

        try:
            r = requests.post(
                mattermostbot_post_url,
                data=json.dumps(payload),
                headers=headers,
            )
            r.raise_for_status()
        except Exception as error:
            logger.exception(f"Error Sending Mattermost Bot message: {error}")
            return False

        return True

    def _notify_mattermostbot(self, act="", message="", tmsg="", force=False):
        if not settings.USE_MATTERMOSTBOT and not force:
            logger.debug("Notification for Mattermost Bot not enabled, skipping this notification")
            return False

        return self._send_mattermostbot(act, message, tmsg)



import math
import socket
import time
from datetime import datetime
from typing import Dict, Iterable, List, TYPE_CHECKING, Union

import jsonrpclib

from sickchill import logger, settings
from sickchill.helper.common import episode_num
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import classes, scene_exceptions, tvcache
from sickchill.oldbeard.common import cpu_presets
from sickchill.oldbeard.helpers import sanitizeSceneName
from sickchill.providers.torrent.TorrentProvider import TorrentProvider

if TYPE_CHECKING:
    from sickchill.tv import TVEpisode


class Provider(TorrentProvider):
    def __init__(self):
        super().__init__("BTN")

        self.supports_absolute_numbering = True

        self.api_key = None

        self.cache = BTNCache(self, min_time=15)  # Only poll BTN every 15 minutes max

        self.urls = {
            "base_url": "https://api.broadcasthe.net",
            "website": "https://broadcasthe.net/",
        }

        self.url = self.urls["website"]

    def _check_auth(self):
        if not self.api_key:
            logger.warning("Invalid api key. Check your settings")

        return True

    def check_auth_from_data(self, data):
        if not data:
            return self._check_auth()

        if "api-error" in data:
            logger.debug(_("Incorrect authentication credentials: {0}").format(data["api-error"]))
            raise AuthException(_("Your authentication credentials for {0} are incorrect, check your config.").format(self.name))

        return True

    def search(self, search_params):
        self._check_auth()

        results = []

        if search_params:
            logger.debug(_("Search string: {0}").format(search_params))
        else:
            # age in seconds
            search_params = {"age": f"<={4 * 24 * 60 * 60}"}

        data = self._api_call(search_params)
        if not data:
            logger.debug("No data returned from provider")
            return results

        found = {}
        if self.check_auth_from_data(data):
            if "torrents" in data:
                found = data["torrents"]

            # We got something, we know the API sends max 1000 results at a time.
            # See if there are more than 1000 results for our query, if not we
            # keep requesting until we've got everything.
            # max 150 requests per hour so limit at that. Scan every 15 minutes. 60 / 15 = 4.
            max_pages = 150
            results_per_page = 1000

            if "results" in data and int(data["results"]) >= results_per_page:
                pages_needed = int(math.ceil(int(data["results"]) / results_per_page))
                if pages_needed > max_pages:
                    pages_needed = max_pages

                # +1 because range(1,4) = 1, 2, 3
                for page in range(1, pages_needed + 1):
                    data = self._api_call(search_params, results_per_page, page * results_per_page)
                    # Note that this these are individual requests and might time out individually. This would result in 'gaps'
                    # in the results. There is no way to fix this though.
                    if "torrents" in data:
                        found.update(data["torrents"])

            for keys, torrent_info in found.items():
                (title, url) = self._get_title_and_url(torrent_info)

                if title and url:
                    logger.debug("Found result: {0} ".format(title))
                    results.append(torrent_info)

        return sorted(results, key=lambda x: self._get_seeders_and_leechers(x)[0], reverse=True)

    def _api_call(self, params=None, results_per_page=1000, offset=0) -> dict:
        data = {}

        try:
            data = jsonrpclib.Server(self.urls["base_url"]).getTorrents(self.api_key, params or {}, int(results_per_page), int(offset))
            time.sleep(cpu_presets[settings.CPU_PRESET])
        except jsonrpclib.jsonrpc.ProtocolError as error:
            if error == (-32001, "Invalid API Key"):
                logger.warning("The API key you provided was rejected because it is invalid. Check your provider configuration.")
            elif error == (-32002, "Call Limit Exceeded"):
                logger.warning("You have exceeded the limit of 150 calls per hour, per API key which is unique to your user account")
            else:
                logger.exception(f"JSON-RPC protocol error while accessing provider. Error: {error} ")
            data = {"api-error": f"{error}"}
            return data

        except socket.timeout:
            logger.warning("Timeout while accessing provider")

        except socket.error as error:
            # Note that sometimes timeouts are thrown as socket errors
            logger.warning(f"Socket error while accessing provider. Error: {error}")

        except Exception as error:
            logger.warning(f"Unknown error while accessing provider. Error: {error}")

        return data

    def _get_seeders_and_leechers(self, item):
        try:
            return item.get("Seeders", -1), item.get("Leechers", -1)
        except AttributeError:
            return -1, -1

    def _get_title_and_url(self, data):
        # The BTN API gives a lot of information in response,
        # however SickChill is built mostly around Scene or
        # release names, which is why we are using them here.
        release_name = data.get("ReleaseName")
        if release_name:
            title = release_name
            append = ".".join(
                f"{part}" for part in (data.get("Resolution"), data.get("Source"), data.get("Codec")) if part and part.lower() not in release_name.lower()
            )
            if append:
                title += " [" + append + "]"

        else:
            # If we don't have a release name we need to get creative
            title = (
                ".".join(
                    f"{part}"
                    for part in (
                        data.get("Series"),
                        data.get("GroupName"),
                        data.get("Resolution"),
                        data.get("Source"),
                        data.get("Codec"),
                    )
                    if part
                )
                .replace(" ", ".")
                .replace("..", ".")
            )

        return title, data.get("DownloadURL", "").replace("\\/", "/") or None

    @staticmethod
    def __add_tvdb_or_name(params, episode):
        search_params = []
        if episode.show.indexer == 1:
            params["tvdb"] = episode.show.indexerid
            search_params.append(params)
        else:
            name_exceptions = list(set(scene_exceptions.get_scene_exceptions(episode.show.indexerid) + [episode.show.name]))
            for name in name_exceptions:
                # Search by name if we don't have tvdb id
                params["series"] = sanitizeSceneName(name)
                search_params.append(params)

        return search_params

    def get_season_search_strings(self, episode: "TVEpisode") -> List[Dict]:
        search_params = {"category": "Season"}

        # Search for entire seasons: no need to do special things for air by date or sports shows
        if self.show.air_by_date or self.show.sports:
            # Search for the year of the air by date show
            search_params["name"] = str(episode.airdate).split("-")[0]
        else:
            # BTN uses the same format for both Anime and TV
            search_params["name"] = "Season " + str(episode.scene_season)

        return self.__add_tvdb_or_name(search_params, episode)

    def get_episode_search_strings(self, episode: "TVEpisode", add_string: str = "") -> Union[List[Dict], Iterable]:
        if not episode:
            return [{}]

        search_params = {"category": "Episode"}

        # episode
        if self.show.air_by_date or self.show.sports:
            date_str = str(episode.airdate)

            # BTN uses dots in dates, we just search for the date since that
            # combined with the series identifier should result in just one episode
            search_params["name"] = date_str.replace("-", ".")
        else:
            # BTN uses the same format for both Anime and TV
            # Do a general name search for the episode
            search_params["name"] = episode_num(episode.scene_season, episode.scene_episode)

        # search
        return self.__add_tvdb_or_name(search_params, episode)

    def find_propers(self, search_date=None):
        results = []

        search_terms = ["%.proper.%", "%.repack.%"]

        for term in search_terms:
            for item in self.search({"release": term}):
                if item["Time"]:
                    try:
                        result_date = datetime.fromtimestamp(float(item["Time"]))
                    except TypeError:
                        result_date = None

                    if result_date and (not search_date or result_date > search_date):
                        title, url = self._get_title_and_url(item)
                        if title and url:
                            results.append(classes.Proper(title, url, result_date, self.show))

        return results


class BTNCache(tvcache.TVCache):
    def _get_rss_data(self):
        return {"entries": self.provider.search(search_params=None)}



import os
import sys
from argparse import ArgumentParser, SUPPRESS


class SickChillArgumentParser:
    def __init__(self, data_dir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_dir = data_dir
        self.parser = ArgumentParser()

        self.parser.description = """SickChill is an automatic tv library manager. It handles searching, sending to your download client, organizing, renaming,
        and adding images and metadata. It handles it all (with a little bit of magic) so you don't have to.
        (c) 2017 SickChill
        """

        self.parser.add_argument("-q", "--quiet", action="store_true", help="disable logging to the console")
        self.parser.add_argument("--nolaunch", action="store_true", help="suppress launching the web browser on startup")
        self.parser.add_argument("-p", "--port", type=int, help="the port to listen on")
        self.parser.add_argument(
            "--datadir",
            help="full path to a folder where the database, config, cache and log files should be stored. Default: {data_dir}"
            "{sep}".format(data_dir=self.data_dir, sep=os.sep),
        )
        self.parser.add_argument(
            "--config",
            help="full file path to override the default configuration file. Default: {data_dir}{sep}config.ini".format(data_dir=self.data_dir, sep=os.sep),
        )
        self.parser.add_argument("--pidfile", help="combined with --daemon creates a pid file (full path)")
        self.parser.add_argument("--noresize", action="store_true", help="prevent resizing of show images even if PIL is installed")
        daemon_help = "run as daemon (includes options --quiet --nolaunch)"
        if sys.platform in ["win32", "darwin"]:
            daemon_help = "running as daemon is not supported on your platform. it is substituted with: --quiet --nolaunch"
        self.parser.add_argument("-d", "--daemon", action="store_true", help=daemon_help)
        self.parser.add_argument(
            "--force-update",
            action="store_true",
            help="download the latest stable version and force an " "update (use when you're unable to do so using " "the web ui)",
        )
        self.parser.add_argument(
            "--install-file",
            help="full path to a wheel (*.whl) or Result.zip from github actions artifact to install. This is for testing other versions of sickchill",
        )

        self.parser.add_argument("--debug", action="store_true", default=False, help=SUPPRESS)
        self.parser.add_argument("--dbdebug", action="store_true", default=False, help=SUPPRESS)
        self.parser.add_argument("--no-file-logging", action="store_true", default=False, help=SUPPRESS)

        self.parser.add_argument("--no-update", action="store_true", help="disable the built-in updater")
        self.parser.add_argument("--flask", action="store_true", help="run the flask server")
        self.parser.add_argument("--flask-host", help="the host for flask listen on", required=False)
        self.parser.add_argument("--flask-port", type=int, help="the port for flask listen on", required=False)

        test_subparser = self.parser.add_subparsers(help="test commands parsers", dest="subparser_name")
        test_name_parser = test_subparser.add_parser("test-name", help="test a release name to see the parse result")
        test_name_parser.add_argument("--name", help="test a release name to see the parse result", required=True)
        test_name_parser.add_argument(
            "--parser",
            help="parser type to use when testing a release name with the parser",
            default="all,normal,anime",
            required=False,
        )

    def parse_args(self, args=None):
        return self.parser.parse_args(args)

    def parse_known_args(self, args=None, namespace=None):
        return self.parser.parse_known_args(args, namespace)



import datetime
import itertools
import re
import time
import traceback
from urllib.parse import urlparse

import validators

from sickchill import logger, settings
from sickchill.helper.common import convert_size, try_int
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.show.Show import Show

from . import db, show_name_helpers
from .databases import cache
from .name_parser.parser import InvalidNameException, InvalidShowException, NameParser

provider_cache_db = {}


class RSSTorrentMixin:
    # {'title': 'Danger.in.the.House.2022.720p.WEB.h264-BAE', 'title_detail': {'type': 'text/plain', 'language': None, 'base': '', 'value': 'Danger.in.the.House.2022.720p.WEB.h264-BAE'}, 'id': 'https://example.com/details/random-guid', 'guidislink': True, 'link': 'https://example.com/getnzb?id=random-guid.nzb&r=1260abc930e0cebbbfc687a42a7c1450', 'links': [{'rel': 'alternate', 'type': 'text/html', 'href': 'https://example.com/getnzb?id=random-guid.nzb&r=1260abc930e0cebbbfc687a42a7c1450'}, {'length': '1962145316', 'type': 'application/x-nzb', 'href': 'https://example.com/getnzb?id=random-guid.nzb&r=1260abc930e0cebbbfc687a42a7c1450', 'rel': 'enclosure'}], 'comments': 'https://example.com/details/random-guid#comments', 'published': 'Tue, 30 Aug 2022 01:30:51 +0200', 'published_parsed': time.struct_time(tm_year=2022, tm_mon=8, tm_mday=29, tm_hour=23, tm_min=30, tm_sec=51, tm_wday=0, tm_yday=241, tm_isdst=0), 'summary': 'Danger.in.the.House.2022.720p.WEB.h264-BAE', 'summary_detail': {'type': 'text/html', 'language': None, 'base': '', 'value': 'Danger.in.the.House.2022.720p.WEB.h264-BAE'}, 'newznab_attr': {'name': 'size', 'value': '1962145316'}}
    #   <item>
    #    <title>Danger.in.the.House.2022.720p.WEB.h264-BAE</title>
    #    <guid isPermaLink="true">https://example.com/details/random-guid</guid>
    #    <link>https://example.com/getnzb?id=random-guid.nzb&amp;r=1260abc930e0cebbbfc687a42a7c1450</link>
    #    <comments>https://example.com/details/random-guid#comments</comments>
    #    <pubDate>Tue, 30 Aug 2022 01:30:51 +0200</pubDate>
    #    <description>Danger.in.the.House.2022.720p.WEB.h264-BAE</description>
    #    <enclosure url="https://example.com/getnzb?id=random-guid.nzb&amp;r=1260abc930e0cebbbfc687a42a7c1450" length="1962145316" type="application/x-nzb"/>
    #    <newznab:attr name="category" value="5010"/>
    #    <newznab:attr name="size" value="1962145316"/>
    #   </item>

    @classmethod
    def check_link(cls, link, url):
        return urlparse(link).netloc == urlparse(url).netloc or validators.url(link) is True or link.startswith("magnet")

    @classmethod
    def parse_feed_item(cls, item, url, size_units=None):
        seeders = 0
        leechers = 0
        item_size = -1
        info_hash = ""
        download_url = None

        title = item.title.get_text(strip=True)

        found_urls = set()
        if item.link:
            found_urls.add(item.link.get_text(strip=True))
            if item.link.next.strip():
                found_urls.add(item.link.next.strip())
        if item.enclosure and item.enclosure.get("url", "").strip():
            found_urls.add(item.enclosure.get("url", "").strip())

        for found_url in found_urls:
            if cls.check_link(found_url, url):
                download_url = found_url

        if not (title and download_url):
            logger.debug(f"Skipping result {title}, {found_urls}")
            logger.debug(f"{item}")
            return

        attribute = item.find(["newznab:attr", "torznab:attr"], attrs={"name": ["infoHash", "info_hash", "guid"]})
        if attribute:
            info_hash = attribute["value"]

        if not info_hash:
            regex = "^.*(?P<guid>[{]?[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}[}]?).*$"
            info_hash = item.find(["infoHash", "info_hash", "guid"]).get_text(strip=True)
            if info_hash:
                match = re.match(regex, info_hash)
                if match:
                    info_hash = match.group("guid")

        attribute = item.find(["newznab:attr", "torznab:attr"], attrs={"name": "seeders"})
        if attribute:
            seeders = try_int(attribute["value"])

        if item.seeders and not seeders:
            seeders = try_int(item.seeders.get_text(strip=True))

        attribute = item.find(["newznab:attr", "torznab:attr"], attrs={"name": ["leechers", "peers"]})
        if attribute:
            leechers = try_int(attribute["value"])

        if not leechers:
            attribute = item.find(["peers", "leechers"])
            if attribute:
                leechers = try_int(attribute.get_text(strip=True))

        attribute = item.find(["newznab:attr", "torznab:attr"], attrs={"name": "size"})
        if attribute:
            item_size = attribute["value"]

        if not item_size:
            if item.size:
                item_size = item.size.get_text(strip=True) or -1
            elif "gingadaddy" in url and item.description:
                size_regex = re.search(r"\d*.?\d* [KMGT]B", item.description.get_text(strip=True))
                if size_regex:
                    item_size = size_regex.group()

        torznab = any([item.seeders, item.leechers, item.peers, download_url.endswith("torrent"), download_url.startswith("magnet")])

        if not item_size or (torznab and (seeders is None or leechers is None)):
            logger.debug(f"Did not find size|seeders|leechers: {item_size}, {seeders}, {leechers}")
            logger.debug(f"{item}")
            return

        if torznab and not seeders:
            # TODO: Implement minseed/minleech for torznab/jackett
            logger.debug(f"Skipping torznab result {title} because there are no seeders.")

        if size_units:
            size = convert_size(item_size, units=size_units) or -1
        else:
            size = convert_size(item_size) or -1

        return {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": info_hash}

    @classmethod
    def check_torznab(cls, soup) -> bool:
        try:
            torznab = soup.find("seeders") or soup.find("server").get("title") == "Jackett"
        except AttributeError:
            try:
                torznab = "xmlns:torznab" in soup.rss.attrs
            except AttributeError:
                torznab = False
        return torznab

    @classmethod
    def getFeed(cls, url, params=None, request_hook=None, size_units=None):
        items = []
        try:
            data = request_hook(url, params=params, returns="text", timeout=30)
            if not data:
                raise Exception

            with BS4Parser(data, language="xml") as feed:
                for item in feed("item"):
                    try:
                        result = cls.parse_feed_item(item, url, size_units=size_units)
                        if result:
                            items.append(result)
                    except Exception as error:
                        logger.debug(f"Error parsing: {error}")
                        logger.debug(traceback.format_exc())
                        continue
        except Exception as error:
            logger.debug(f"RSS error: {error}")

        return {"entries": items}


class CacheDBConnection(db.DBConnection):
    def __init__(self):
        super().__init__("cache.db")
        db.upgrade_database(self, cache.InitialSchema)
        # self.action("DELETE from results WHERE added < datetime('now','-30 days')")

        self.trim()

    def trim(self, provider: str = ""):
        sql = f"DELETE from results WHERE added < datetime('now','-{settings.CACHE_RETENTION} days')"
        sql_args = []
        if provider:
            sql += "AND provider = ?"
            sql_args.append(provider)
        self.action(sql, sql_args)


class TVCache(RSSTorrentMixin):
    def __init__(self, provider, **kwargs):
        self.provider = provider
        self.provider_id = self.provider.get_id()
        self.provider_db = None
        self.min_time = kwargs.pop("min_time", 10)
        self.search_params = kwargs.pop("search_params", dict(RSS=[""]))

    def get_db(self):
        # init provider database if not done already
        if not provider_cache_db.get("instance"):
            provider_cache_db["instance"] = CacheDBConnection()

        return provider_cache_db.get("instance")

    def _clear_cache(self):
        if self.should_clear_cache():
            cache_db_con = self.get_db()
            cache_db_con.trim(self.provider_id)

    def _get_seeders_and_leechers(self, item):
        return self.provider._get_seeders_and_leechers(item)

    def _get_size(self, item):
        return self.provider._get_size(item)

    def _get_title_and_url(self, item):
        return self.provider._get_title_and_url(item)

    def _get_rss_data(self):
        return {"entries": self.provider.search(self.search_params)} if self.search_params else None

    def _check_auth(self, data):
        return True

    def update_cache(self):
        # check if we should update
        if not self.should_update():
            return

        try:
            data = self._get_rss_data()
            if self._check_auth(data):
                # clear cache
                self._clear_cache()

                # set updated
                self.set_last_update()

                cl = []
                for item in data["entries"] or []:
                    ci = self._parse_item(item)
                    if ci:
                        cl.append(ci)

                if cl:
                    cache_db_con = self.get_db()
                    cache_db_con.mass_upsert("results", cl)

        except AuthException as error:
            logger.warning(f"Authentication error: {error}")
        except Exception as error:
            logger.debug(f"Error while searching {self.provider.name}, skipping: {error}")
            logger.debug(traceback.format_exc())

    def get_rss_feed(self, url, params=None):
        if self.provider.login():
            return self.getFeed(url, params=params, request_hook=self.provider.get_url, size_units=self.provider.size_units)
        return {"entries": []}

    @staticmethod
    def _translate_title(title):
        return "" + title.replace(" ", ".")

    @staticmethod
    def _translate_link_url(url):
        return url.replace("&amp;", "&")

    def _parse_item(self, item):
        title, url = self._get_title_and_url(item)
        size = self._get_size(item)
        seeders, leechers = self._get_seeders_and_leechers(item)

        if title and url:
            title = self._translate_title(title)
            url = self._translate_link_url(url)

            # logger.debug("Attempting to add item to cache: " + title)
            return self.add_cache_entry(title, url, size, seeders, leechers)

        else:
            logger.debug("The data returned from the " + self.provider.name + " feed is incomplete, this result is unusable")

        return False

    @property
    def last_update(self):
        cache_db_con = self.get_db()
        sql_results = cache_db_con.select("SELECT time FROM lastUpdate WHERE provider = ?", [self.provider_id])

        if sql_results:
            last_time = int(sql_results[0]["time"])
            if last_time > int(time.mktime(datetime.datetime.today().timetuple())):
                last_time = 0
        else:
            last_time = 0

        return datetime.datetime.fromtimestamp(last_time)

    @property
    def last_search(self):
        cache_db_con = self.get_db()
        sql_results = cache_db_con.select("SELECT time FROM lastSearch WHERE provider = ?", [self.provider_id])

        if sql_results:
            last_time = int(sql_results[0]["time"])
            if last_time > int(time.mktime(datetime.datetime.today().timetuple())):
                last_time = 0
        else:
            last_time = 0

        return datetime.datetime.fromtimestamp(last_time)

    def set_last_update(self, to_date=None):
        """
        Sets the last update date for the current provider in the cache database

        :param to_date: date to set to, or None for today
        """
        if not to_date:
            to_date = datetime.datetime.today()

        cache_db_con = self.get_db()
        cache_db_con.upsert("lastUpdate", {"time": int(time.mktime(to_date.timetuple()))}, {"provider": self.provider_id})

    def set_last_search(self, to_date=None):
        """
        Sets the last search date for the current provider in the cache database

        :param to_date: date to set to, or None for today
        """
        if not to_date:
            to_date = datetime.datetime.today()

        cache_db_con = self.get_db()
        cache_db_con.upsert("lastSearch", {"time": int(time.mktime(to_date.timetuple()))}, {"provider": self.provider_id})

    def should_update(self):
        # if we've updated recently then skip the update
        if datetime.datetime.today() - self.last_update < datetime.timedelta(minutes=self.min_time):
            logger.debug("Last update was too soon, using old cache: " + str(self.last_update) + ". Updated less then " + str(self.min_time) + " minutes ago")
            return False

        return True

    def should_clear_cache(self):
        # if daily search hasn't used our previous results yet then don't clear the cache
        if self.last_update > self.last_search:
            return False

        return True

    def add_cache_entry(self, name, url, size, seeders, leechers, parse_result=None, indexer_id=0):
        # check if we passed in a parsed result or should we try and create one
        if not parse_result:
            # create show_obj from indexer_id if available
            show_obj = None
            if indexer_id:
                show_obj = Show.find(settings.show_list, indexer_id)

            try:
                parse_result = NameParser(show_object=show_obj).parse(name)
            except (InvalidNameException, InvalidShowException) as error:
                logger.debug(f"{error}")
                return None

            if not parse_result or not parse_result.series_name:
                return None

        # if we made it this far then lets add the parsed result to cache for usage later on
        season = parse_result.season_number if parse_result.season_number else 1
        episodes = parse_result.episode_numbers

        if season and episodes:
            # store episodes as a separated string
            episode_text = "|" + "|".join({str(episode) for episode in sorted(episodes) if episode}) + "|"

            # get the current timestamp
            cur_timestamp = int(time.mktime(datetime.datetime.today().timetuple()))

            # get quality of release
            quality = parse_result.quality

            # get release group
            release_group = parse_result.release_group

            # get version
            version = parse_result.version

            logger.debug(_("Added RSS item: [{name}] to cache: {provider}").format(name=name, provider=self.provider_id))
            return (
                {
                    "provider": self.provider_id,
                    "name": name,
                    "season": season,
                    "episodes": episode_text,
                    "indexerid": parse_result.show.indexerid,
                    "url": url,
                    "time": cur_timestamp,
                    "quality": quality,
                    "release_group": release_group,
                    "version": version,
                    "seeders": seeders,
                    "leechers": leechers,
                    "size": size,
                },
                {"url": url},
            )

    def search_cache(self, episode, manual_search=False, down_cur_quality=False):
        needed_eps = self.find_needed_episodes(episode, manual_search, down_cur_quality)
        return needed_eps.get(episode, [])

    def list_propers(self, date=None):
        cache_db_con = self.get_db()
        sql = "SELECT * FROM results WHERE provider = ? AND name LIKE '%.PROPER.%' OR name LIKE '%.REPACK.%'"
        # Add specific provider proper_strings also, like REAL, RERIP, etc.
        if hasattr(self.provider, "proper_strings"):
            if self.provider.proper_strings:
                for item in self.provider.proper_strings:
                    if "|" in item:
                        items = item.split("|")
                        for _item in items:
                            if _item.upper() not in sql:
                                sql += " OR name LIKE '%.{}.%'".format(_item)
                    elif item.upper() not in sql:
                        sql += " OR name LIKE '%.{}.%'".format(item)

        if date is not None:
            sql += " AND time >= " + str(int(time.mktime(date.timetuple())))

        propers_results = cache_db_con.select(sql, [self.provider_id])
        return [x for x in propers_results if x["indexerid"]]

    def find_needed_episodes(self, episode, manual_search=False, down_cur_quality=False):
        needed_eps = {}
        cl = []

        cache_db_con = self.get_db()
        if not episode:
            sql_results = cache_db_con.select("SELECT * FROM results WHERE provider = ?", [self.provider_id])
        elif not isinstance(episode, list):
            sql_results = cache_db_con.select(
                "SELECT * FROM results WHERE provider = ? AND indexerid = ? AND season = ? AND episodes LIKE ?",
                [self.provider_id, episode.show.indexerid, episode.season, "%|" + str(episode.episode) + "|%"],
            )
        else:
            for episode_object in episode:
                cl.append(
                    [
                        "SELECT * FROM results WHERE provider = ? AND indexerid = ? AND season = ? AND episodes LIKE ? AND quality IN ("
                        + ",".join([str(x) for x in episode_object.wantedQuality])
                        + ")",
                        [self.provider_id, episode_object.show.indexerid, episode_object.season, "%|" + str(episode_object.episode) + "|%"],
                    ]
                )

            sql_results = cache_db_con.mass_action(cl, fetchall=True)
            sql_results = list(itertools.chain(*sql_results))

        # for each cache entry
        for cur_result in sql_results:
            # get the show object, or if it's not one of our shows then ignore it
            show_obj = Show.find(settings.show_list, int(cur_result["indexerid"]))
            if not show_obj:
                continue

            # ignored/required words, and non-tv junk
            if not show_name_helpers.filter_bad_releases(cur_result["name"], show=show_obj):
                continue

            # skip if provider is anime only and show is not anime
            if self.provider.anime_only and not show_obj.is_anime:
                logger.debug("" + str(show_obj.name) + " is not an anime, skiping")
                continue

            # get season and ep data (ignoring multi-eps for now)
            cur_season = int(cur_result["season"])
            if cur_season == -1:
                continue

            cur_ep = cur_result["episodes"].split("|")[1]
            if not cur_ep:
                continue

            cur_ep = int(cur_ep)

            cur_quality = int(cur_result["quality"])
            cur_release_group = cur_result["release_group"]
            cur_version = cur_result["version"]

            # if the show says we want that episode then add it to the list
            if not show_obj.want_episode(cur_season, cur_ep, cur_quality, manual_search, down_cur_quality):
                logger.debug("Ignoring " + cur_result["name"])
                continue

            episode_object = show_obj.get_episode(cur_season, cur_ep)

            # build a result object
            title = cur_result["name"]
            url = cur_result["url"]

            logger.info("Found result " + title + " at " + url)

            result = self.provider.get_result([episode_object], url)
            result.show = show_obj
            result.name = title
            result.quality = cur_quality
            result.release_group = cur_release_group
            result.version = cur_version
            result.content = None

            # add it to the list
            if episode_object not in needed_eps:
                needed_eps[episode_object] = [result]
            else:
                needed_eps[episode_object].append(result)

        # datetime stamp this search so cache gets cleared
        self.set_last_search()

        return needed_eps



import datetime
import glob
import os
import re
import time

from sickchill import logger, settings
from sickchill.helper.exceptions import UpdaterException
from sickchill.oldbeard import db, helpers, ui

from .pip import PipUpdateManager


class UpdateManager(object):
    """
    Version check class meant to run as a thread object with the sc scheduler.
    """

    def __init__(self):
        self.updater = None
        self.amActive = False
        self.updater = None

        if not settings.DISABLE_UPDATER:
            self.updater = PipUpdateManager()

        self.session = helpers.make_session()

    def run(self, force=False):
        self.amActive = True

        if self.updater:
            if self.check_for_new_version(force):
                if settings.AUTO_UPDATE:
                    logger.info("New update found for SickChill, starting auto-updater ...")
                    ui.notifications.message(_("New update found for SickChill, starting auto-updater"))
                    if self.run_backup_if_safe():
                        if settings.versionCheckScheduler.action.update():
                            logger.info("Update was successful!")
                            ui.notifications.message(_("Update was successful"))
                            settings.events.put(settings.events.SystemEvent.RESTART)
                        else:
                            logger.info("Update failed!")
                            ui.notifications.message(_("Update failed!"))

        self.check_for_new_news()

        self.amActive = False

    def run_backup_if_safe(self):
        return self.safe_to_update() is True and self._run_backup() is True

    def _run_backup(self):
        # Do a system backup before update
        logger.info("Config backup in progress...")
        ui.notifications.message(_("Backup"), _("Config backup in progress..."))
        try:
            backup_dir = os.path.join(settings.DATA_DIR, "backup")
            if not os.path.isdir(backup_dir):
                os.mkdir(backup_dir)

            if self._keep_latest_backup(backup_dir) and self.backup_to_dir(backup_dir):
                logger.info("Config backup successful, updating...")
                ui.notifications.message(_("Backup"), _("Config backup successful, updating..."))
                return True
            else:
                logger.exception("Config backup failed, aborting update")
                ui.notifications.message(_("Backup"), _("Config backup failed, aborting update"))
                return False
        except Exception as error:
            logger.exception(f"Update: Config backup failed. Error: {error}")
            ui.notifications.message(_("Backup"), _("Config backup failed, aborting update"))
            return False

    @staticmethod
    def _keep_latest_backup(backup_dir=None):
        if not backup_dir:
            return False

        files = glob.glob(os.path.join(glob.escape(backup_dir), "*.zip"))
        if not files:
            return True

        now = time.time()
        newest = files[0], now - os.path.getctime(files[0])
        for f in files[1:]:
            age = now - os.path.getctime(f)
            if age < newest[1]:
                newest = f, age
        files.remove(newest[0])

        for f in files:
            os.remove(f)

        return True

    @staticmethod
    def backup_to_dir(backup_dir=None):
        if not backup_dir:
            return False
        source = [
            os.path.join(settings.DATA_DIR, "sickchill.db"),
            settings.CONFIG_FILE,
            os.path.join(settings.DATA_DIR, "failed.db"),
            os.path.join(settings.DATA_DIR, "cache.db"),
        ]
        target = os.path.join(backup_dir, "sickchill-" + time.strftime("%Y%m%d%H%M%S") + ".zip")

        for path, dirs, files in os.walk(settings.CACHE_DIR):
            for dirname in dirs:
                if path == settings.CACHE_DIR and dirname not in ["images"]:
                    dirs.remove(dirname)
            for filename in files:
                source.append(os.path.join(path, filename))

        return helpers.backup_config_zip(source, target, settings.DATA_DIR)

    def safe_to_update(self):
        def db_safe():
            message = {
                "equal": {"type": logger.DEBUG, "text": "We can proceed with the update. New update has same DB version"},
                "upgrade": {"type": logger.WARNING, "text": "We can't proceed with the update. New update has a new DB version. Please manually update"},
                "downgrade": {
                    "type": logger.ERROR,
                    "text": "We can't proceed with the update. New update has a old DB version. It's not possible to downgrade",
                },
            }
            try:
                result = self.compare_db_version()
                if result in message:
                    logger.log(message[result]["type"], message[result]["text"])  # unpack the result message into a log entry
                else:
                    logger.warning("We can't proceed with the update. Unable to check remote DB version. Error: {0}".format(result))
                return result in ["equal"]  # add future True results to the list
            except Exception as error:
                logger.warning(f"We can't proceed with the update. Unable to compare DB version. Error: {error}")
                return False

        def postprocessor_safe():
            if not settings.autoPostProcessorScheduler.action.amActive:
                logger.debug("We can proceed with the update. Post-Processor is not running")
                return True
            else:
                logger.debug("We can't proceed with the update. Post-Processor is running")
                return False

        def showupdate_safe():
            if not settings.showUpdateScheduler.action.amActive:
                logger.debug("We can proceed with the update. Shows are not being updated")
                return True
            else:
                logger.debug("We can't proceed with the update. Shows are being updated")
                return False

        db_safe = db_safe()
        postprocessor_safe = postprocessor_safe()
        showupdate_safe = showupdate_safe()

        if db_safe and postprocessor_safe and showupdate_safe:
            logger.debug("Proceeding with auto update")
            return True
        else:
            logger.debug("Auto update aborted")
            return False

    def compare_db_version(self):
        try:
            self.need_update()
            newest_version = self.get_newest_version()
            response = helpers.getURL(
                f"https://raw.githubusercontent.com/{settings.GIT_ORG}/{settings.GIT_REPO}/{newest_version}/sickchill/oldbeard/databases/main.py",
                session=self.session,
                returns="text",
            )
            if not response:
                response = helpers.getURL(
                    f"https://raw.githubusercontent.com/{settings.GIT_ORG}/{settings.GIT_REPO}/master/sickchill/oldbeard/databases/main.py",
                    session=self.session,
                    returns="text",
                )

            if not response:
                raise UpdaterException(f"Empty response from GitHub for {newest_version}")

            match = re.search(r"MAX_DB_VERSION\s=\s(?P<version>\d{2,3})", response)
            destination_db_version = int(match.group("version"))
            main_db_con = db.DBConnection()
            current_db_version = main_db_con.get_db_version()
            if destination_db_version > current_db_version:
                return "upgrade"
            elif destination_db_version == current_db_version:
                return "equal"
            else:
                return "downgrade"
        except Exception as error:
            return f"{error}"

    def check_for_new_version(self, force=False):
        """
        Checks the internet for a newer version.

        returns: bool, True for new version or False for no new version.

        force: if true the VERSION_NOTIFY setting will be ignored and a check will be forced
        """
        if not self.updater or (not settings.VERSION_NOTIFY and not settings.AUTO_UPDATE and not force):
            logger.info("Version checking is disabled, not checking for the newest version")
            return False

        # checking for updates
        if not settings.AUTO_UPDATE:
            logger.info("Checking for updates from pip")

        if not self.need_update():
            if force:
                ui.notifications.message(_("No update needed"))
                logger.info("No update needed")

            # no updates needed
            return False

        # found updates
        self.set_newest_text()
        return True

    def check_for_new_news(self):
        """
        Checks GitHub for the latest news.

        returns: str, a copy of the news

        force: ignored
        """

        # Grab a copy of the news
        logger.debug("check_for_new_news: Checking GitHub for latest news.")
        try:
            news = helpers.getURL(settings.NEWS_URL, session=self.session, returns="text")
        except Exception:
            logger.warning("check_for_new_news: Could not load news from repo.")
            news = ""

        if not news:
            return ""

        try:
            last_read = datetime.datetime.strptime(settings.NEWS_LAST_READ, "%Y-%m-%d")
        except Exception:
            last_read = 0

        settings.NEWS_UNREAD = 0
        found_news = False
        for match in re.finditer(r"^####\s*(\d{4}-\d{2}-\d{2})\s*####", news, re.M):
            if not found_news:
                found_news = True
                settings.NEWS_LATEST = match.group(1)

            try:
                if datetime.datetime.strptime(match.group(1), "%Y-%m-%d") > last_read:
                    settings.NEWS_UNREAD += 1
            except Exception:
                pass

        return news

    def update(self):
        if self.need_update():
            return self.updater.update()

    def need_update(self):
        if self.updater:
            return self.updater.need_update()

    def backup(self):
        if self.updater:
            return self._run_backup

    def get_newest_version(self):
        if self.updater:
            return self.updater.get_newest_version()

    def get_current_version(self):
        if self.updater:
            return self.updater.get_current_version()

    def get_version_delta(self):
        if self.updater:
            return self.updater.get_version_delta()

    def set_newest_text(self):
        if self.updater:
            return self.updater.set_newest_text()



import datetime
import http.client
import xmlrpc.client
from base64 import standard_b64encode
from typing import TYPE_CHECKING

from sickchill import logger, settings
from sickchill.helper.common import try_int
from sickchill.oldbeard.helpers import make_context

from .common import Quality

if TYPE_CHECKING:
    from .classes import SearchResult


def get_proxy(https: bool, host: str, username: str, password: str, verify: bool) -> xmlrpc.client.ServerProxy:
    scheme = "http"
    if https:
        scheme += "s"
    return xmlrpc.client.ServerProxy(f"{scheme}://{username}:{password}@{host}/xmlrpc", context=make_context(verify))


def send_nzb(result: "SearchResult", proper=False) -> bool:
    """
    Sends NZB to NZBGet client

    :param result: search result
    :param proper: True if this is a Proper download, False if not. Defaults to False
    """
    if not settings.NZBGET_HOST:
        logger.warning("No NZBget host found in configuration. Please configure it.")
        return False

    proxy = get_proxy(settings.NZBGET_USE_HTTPS, settings.NZBGET_HOST, settings.NZBGET_USERNAME, settings.NZBGET_PASSWORD, settings.SSL_VERIFY)
    try:
        if proxy.writelog("INFO", _("SickChill connected to drop off {name} any moment now.").format(name=f"{result.name}.nzb")):
            logger.debug("Successfully connected to NZBget")
        else:
            logger.warning("Successfully connected to NZBget, but unable to send a message")

    except http.client.error:
        logger.warning("Please check your NZBget host and port (if it is running). NZBget is not responding to this combination")
        return False

    except xmlrpc.client.ProtocolError as error:
        if error.errmsg == "Unauthorized":
            logger.warning("NZBget username or password is incorrect.")
        else:
            logger.exception(f"Protocol Error: {error}")
        return False

    newest_episode_age = datetime.date.today() - max({episode.airdate for episode in result.episodes if episode.airdate})
    if newest_episode_age <= datetime.timedelta(days=7):
        add_to_top = True
        nzbget_priority = settings.NZBGET_PRIORITY

        category = settings.NZBGET_CATEGORY
        if result.show.is_anime:
            category = settings.NZBGET_CATEGORY_ANIME

    else:
        add_to_top = False
        nzbget_priority = 0

        category = settings.NZBGET_CATEGORY_BACKLOG
        if result.show.is_anime:
            category = settings.NZBGET_CATEGORY_ANIME_BACKLOG

    dupe_key = ""
    # if it aired recently make it high priority and generate DupeKey/Score
    for curEp in result.episodes:
        if not dupe_key:
            dupe_key = f"SickChill-{curEp.show.indexerid}"
        dupe_key += f"-{curEp.season:02d}{curEp.episode:02d}"

    dupe_score = result.quality or 0

    if result.show.quality and dupe_score:
        allowed_qualities, preferred_qualities = Quality.splitQuality(result.show.quality)
        if preferred_qualities:
            if result.quality == max(preferred_qualities):
                dupe_score *= 1000
            elif result.quality in preferred_qualities:
                dupe_score *= 800
        elif allowed_qualities:
            if result.quality == max(allowed_qualities):
                dupe_score *= 500
            elif result.quality in allowed_qualities:
                dupe_score *= 300

    elif dupe_score and dupe_score != Quality.UNKNOWN:
        dupe_score *= 100

    if proper:
        dupe_score += 10

    nzb_data_content = None
    if result.is_nzbdata:
        data = result.extraInfo[0]
        nzb_data_content = standard_b64encode(data)

    logger.info("Sending NZB to NZBget")
    logger.debug(f"URL: {result.url}")

    try:
        # Find out if nzbget supports priority (Version 9.0+), old versions beginning with a 0.x will use the old command
        nzbget_version_str = proxy.version()
        nzbget_version = try_int(nzbget_version_str[: nzbget_version_str.find(".")])
        if nzbget_version == 0:
            if nzb_data_content:
                nzbget_result = proxy.append(f"{result.name}.nzb", category, add_to_top, nzb_data_content)
            else:
                if result.is_nzb:
                    if not result.provider.login():
                        return False

                    data = result.provider.get_url(result.url, returns="content")
                    if not data:
                        return False

                    nzb_data_content = standard_b64encode(data)

                nzbget_result = proxy.append(f"{result.name}.nzb", category, add_to_top, nzb_data_content)
        elif nzbget_version == 12:
            if nzb_data_content is not None:
                nzbget_result = proxy.append(f"{result.name}.nzb", category, nzbget_priority, False, nzb_data_content, False, dupe_key, dupe_score, "score")
            else:
                nzbget_result = proxy.appendurl(f"{result.name}.nzb", category, nzbget_priority, False, result.url, False, dupe_key, dupe_score, "score")
        # v13+ has a new combined append method that accepts both (url and content)
        # also the return value has changed from boolean to integer
        # (Positive number representing NZBID of the queue item. 0 and negative numbers represent error codes.)
        elif nzbget_version >= 13:
            nzbget_result = (
                proxy.append(
                    f"{result.name}.nzb",
                    nzb_data_content if nzb_data_content is not None else result.url,
                    category,
                    nzbget_priority,
                    False,
                    False,
                    dupe_key,
                    dupe_score,
                    "score",
                )
                > 0
            )
        else:
            if nzb_data_content is not None:
                nzbget_result = proxy.append(f"{result.name}.nzb", category, nzbget_priority, False, nzb_data_content)
            else:
                nzbget_result = proxy.appendurl(f"{result.name}.nzb", category, nzbget_priority, False, result.url)

        if nzbget_result:
            logger.debug("NZB sent to NZBget successfully")
            return True
        else:
            logger.warning(_("NZBget could not add {name} to the queue").format(name=f"{result.name}.nzb"))
            return False
    except Exception:
        logger.warning(_("Connect Error to NZBget: could not add {name} to the queue").format(name=f"{result.name}.nzb"))
        return False


def test_client_connection(https: bool, host: str, username: str, password: str, verify: bool):
    proxy = get_proxy(https, host, username, password, verify)

    try:
        if proxy.writelog("INFO", "SickChill connection test succeeded."):
            logger.debug("NZBget connections succeeded")
        else:
            logger.warning("NZBget connections succeeded, but unable to send a message")

        return True
    except http.client.error:
        logger.warning("NZBGet connection test failed, check your NZBget host and port (if it is running). NZBget is not responding to this combination")
        return False



import logging
import tempfile
from pathlib import Path
from unittest import skip, TestCase
from unittest.mock import patch

from sickchill.helper.rootdirs import RootDirectories

logging.basicConfig(format="{levelname} :: {message}", style="{", level=logging.DEBUG)


@skip("Not implemented")
class TestRootDirectories(TestCase):
    path_one = None
    path_two = None
    patcher = None
    settings_root_dirs = None

    @classmethod
    def setUpClass(cls):
        cls.path_one = Path(__file__).parent.resolve()
        cls.path_two = Path(__file__).parent.parent.resolve()

        cls.settings_root_dirs = f"1|{cls.path_one}|{cls.path_two}"

        cls.patcher = patch("sickchill.settings.ROOT_DIRS", cls.settings_root_dirs)
        cls.patcher.start()
        cls.addClassCleanup(cls.patcher.stop)

    def setUp(self):
        self.object = RootDirectories()
        self.initial = str(self.object)
        self.initial_length = len(self.object)
        self.initial_list = [item for item in self.object]

        self.initial_index = self.object.default_root_index

    def test_parse(self):
        from sickchill import settings

        self.object.parse()
        self.assertEqual(self.initial, str(self.object))

        self.object.parse(settings.ROOT_DIRS)
        self.assertEqual(self.initial, str(self.object))

        self.object.root_directories.clear()
        self.object.default_root_index = None

        self.object.parse(settings.ROOT_DIRS)
        self.assertEqual(self.initial, str(self.object))
        self.assertEqual(len(self.object.root_directories), self.initial_length)
        self.assertEqual(self.object.root_directories, self.initial_list)
        self.assertEqual(self.object.default_root_index, self.initial_index)

        self.object.parse(root_directories_list=settings.ROOT_DIRS.split("|"))
        self.assertEqual(len(self.object.root_directories), self.initial_length)
        self.assertEqual(self.object.root_directories, self.initial_list)
        self.assertEqual(self.object.default_root_index, self.initial_index)
        self.assertEqual(self.object.root_directories, RootDirectories(root_directories_string=self.settings_root_dirs).root_directories)

    def test_add(self):
        self.object.add(self.path_one)
        self.assertEqual(len(self.object.root_directories), self.initial_length, msg=f"{self.object.root_directories}")
        self.assertEqual(self.object.root_directories, self.initial_list)
        self.assertEqual(self.object.default_root_index, self.initial_index)

        self.object.add(tempfile.tempdir)

        self.assertNotEqual(len(self.object.root_directories), self.initial_length)
        self.assertNotEqual(self.object.root_directories, self.initial_list)
        self.assertEqual(self.object.default_root_index, self.initial_index)

    def test_delete(self):
        self.object.delete(self.path_one)
        self.assertLess(len(self.object), self.initial_length)

        length = len(self.object)
        self.object.delete(self.path_one)
        self.assertEqual(len(self.object), length)
        self.assertIsNotNone(self.object.default_root_index)
        self.assertEqual(self.object.default_root_index, self.initial_index)

        self.object.delete(self.path_two)
        self.assertFalse(len(self.object))

        self.assertIsNone(self.object.default_root_index)

    def test_default(self):
        self.object.add(tempfile.tempdir)
        self.assertEqual(self.object.default, self.initial_index)

        self.object.default = 10
        self.assertEqual(self.object.default, self.initial_index)
        self.assertEqual(self.object.default, self.initial_index)

        self.object.delete(self.path_one)
        self.object.delete(self.path_two)
        self.object.delete(tempfile.tempdir)
        self.assertIsNone(self.object.default, self.object.root_directories)

        self.object.add(tempfile.tempdir)
        self.assertEqual(self.object.default, self.initial_index)

    def test_info(self):
        info = self.object.info()
        self.assertEqual(len(info), len(self.object))

        self.assertIn("valid", info[0])
        self.assertEqual(info[0]["valid"], True)

        self.assertIn("default", info[0])
        self.assertEqual(info[0]["default"], True)

        self.assertEqual(info[1]["default"], False)
        self.assertEqual(info[0]["location"], str(self.path_one))

        self.object.delete(self.path_one)
        self.object.delete(self.path_two)

        info = self.object.info()
        self.assertEqual(len(info), len(self.object))
        self.assertEqual(info, {})

    def test_str(self):
        self.assertEqual(self.object.__str__(), self.settings_root_dirs)

        self.object.add(tempfile.tempdir)
        self.assertNotEqual(self.object.__str__(), self.settings_root_dirs)

        self.object.delete(tempfile.tempdir)
        self.assertEqual(self.object.__str__(), self.settings_root_dirs)

    def test_getitem(self):
        self.assertEqual(self.object[1], self.path_one)
        self.assertEqual(self.object[1], self.object.root_directories[0])

    def test_setitem(self):
        self.assertEqual(self.object[1], self.path_one)
        self.object[1] = self.path_two
        self.assertEqual(self.object[1], self.path_two)

    def test_del(self):
        del self.object[1]
        self.assertEqual(self.object[1], self.path_two)
        self.assertLess(len(self.object), self.initial_length)

    def test_contains(self):
        self.assertFalse(tempfile.tempdir in self.object)
        self.assertTrue(self.path_one in self.object)
        self.assertTrue(self.path_two in self.object)

        self.object.delete(self.path_one)
        self.assertFalse(self.path_one in self.object)
        self.assertTrue(self.path_two in self.object)

    def test_iter(self):
        object_list = []
        for item in self.object:
            object_list.append(item)

        self.assertEqual(object_list, self.object.root_directories)

    def test_len(self):
        self.assertEqual(len(self.object), len(self.object.root_directories))



from pathlib import Path
from typing import Union

from sickchill import logger, settings


class RootDirectories:
    def __init__(self, root_directories_string: str = "", root_directories_list: Union[list, None] = None) -> None:
        self.default_root_index: Union[int, None] = None
        self.root_directories: list = []

        self.parse(root_directories_string or settings.ROOT_DIRS, root_directories_list)
        logger.debug(f"parsed roots: {[str(directory) for directory in self]} from {settings.ROOT_DIRS}")
        if self.default:
            logger.debug(f"default root directory: {self[self.default]}")
        else:
            logger.debug(f"no default root directory set")

    def parse(self, root_directories_string: str = "", root_directories_list: Union[list, None] = None) -> None:
        split_setting = root_directories_list or root_directories_string.split("|")

        if len(split_setting) < 2:
            return

        self.default_root_index = int(split_setting[0])
        for root_directory in split_setting[1:]:
            self.add(root_directory)

    def add(self, root_directory: Union[str, Path]) -> bool:
        root_directory = Path(root_directory).resolve()
        if not root_directory.is_dir():
            logger.debug(f"tried to add a non-existent root directory: {root_directory}")
            return False

        if root_directory in self.root_directories:
            logger.debug(f"tried to add a root directory that is already added: {root_directory}")
            return False

        logger.debug(f"adding root directory {root_directory}")
        self.root_directories.append(root_directory)

        self.__check_default_index()

        return True

    def delete(self, root_directory: Union[str, Path]) -> bool:
        root_directory = Path(root_directory)
        if root_directory in self.root_directories:
            self.root_directories.remove(root_directory)
            self.__check_default_index()
            return True
        return False

    @property
    def default(self) -> int:
        return self.default_root_index

    @default.setter
    def default(self, value: Union[int, str]) -> None:
        value = int(value)
        self.default_root_index = value
        self.__check_default_index()

    def info(self) -> list:
        output = []
        for root_dir in self:
            output.append({"valid": root_dir.is_dir(), "location": str(root_dir), "default": int(root_dir is self[self.default])})

        return output or {}

    def __check_default_index(self):
        if not self.root_directories:
            self.default_root_index = None
            return

        if not self.default_root_index or self.default_root_index > len(self.root_directories):
            self.default_root_index = 1

    def __str__(self) -> str:
        return "|".join(f"{item}" for item in [self.default_root_index] + self.root_directories)

    def __getitem__(self, item: int) -> Path:
        return self.root_directories[item - 1]

    def __setitem__(self, index: int, value: Path):
        self.root_directories[index - 1] = value
        self.__check_default_index()

    def __delitem__(self, key):
        self.root_directories.pop(key - 1)
        self.__check_default_index()

    def __contains__(self, item):
        return item in self.root_directories

    def __iter__(self):
        for item in self.root_directories:
            yield item

    def __len__(self):
        return len(self.root_directories)



import re

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.FrenchProvider import FrenchTorrentProvider


class Provider(FrenchTorrentProvider):
    def __init__(self):
        super().__init__("Torrent911", "https://www.torrent911.ph")

    def search(self, search_strings):
        results = []
        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}").format(mode=mode))
            for search_string in {*search_strings[mode]}:
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}").format(search_string=search_string))

                    search_url = self.url + "/recherche/" + search_string
                else:
                    search_url = self.url + "/torrents/series/date/desc"

                data = self.get_url(search_url, returns="text")
                if not data:
                    continue

                with BS4Parser(data) as html:
                    # Skip column headers
                    for result in html.select("table.table-hover tr")[1:]:
                        try:
                            title = result.select_one(".maxi").get_text(strip=True).replace("HDTV", "HDTV x264-Torrent911")
                            title = re.sub(r" Saison", " Season", title, flags=re.I)
                            download_url = self._retrieve_dllink_from_url(result.select_one("a").get("href"))
                            if not all([title, download_url]):
                                logger.debug(_("Could not find title and download url for result"))
                                continue

                            seeders = try_int(result.find(attrs={"src": re.compile(r".*.uploader\.png")}).parent.get_text(strip=True))
                            leechers = try_int(result.find(attrs={"src": re.compile(r".*.downloader\.png")}).parent.get_text(strip=True))
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        f"Discarding torrent because it doesn't meet the minimum seeders or leechers: {title} (S:{seeders} L:{leechers})"
                                    )
                                continue

                            size = convert_size(result.find("td", text=re.compile(r"[\d0-9]* [KMG]B")).get_text(strip=True), -1) or -1

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug(f"Found result: {title} with {seeders} seeders and {leechers} leechers")

                            items.append(item)
                        except Exception as error:
                            logger.debug(f"Error parsing results {error}")
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



import datetime
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sickchill.tv import TVEpisode, TVShow

from sickchill.oldbeard import helpers

from ... import logger
from . import generic


class TIVOMetadata(generic.GenericMetadata):
    """
    Metadata generation class for TIVO

    The following file structure is used:

    show_root/Season ##/filename.ext            (*)
    show_root/Season ##/.meta/filename.ext.txt  (episode metadata)

    This class only generates episode specific metadata files, it does NOT generate a default.txt file.
    """

    def __init__(
        self,
        show_metadata=False,
        episode_metadata=False,
        fanart=False,
        poster=False,
        banner=False,
        episode_thumbnails=False,
        season_posters=False,
        season_banners=False,
        season_all_poster=False,
        season_all_banner=False,
    ):
        super().__init__(
            show_metadata, episode_metadata, fanart, poster, banner, episode_thumbnails, season_posters, season_banners, season_all_poster, season_all_banner
        )

        self.name = "TIVO"

        self._ep_nfo_extension = "txt"

        # web-ui metadata template
        self.eg_show_metadata = "<i>not supported</i>"
        self.eg_episode_metadata = "Season##\\.meta\\<i>filename</i>.ext.txt"
        self.eg_fanart = "<i>not supported</i>"
        self.eg_poster = "<i>not supported</i>"
        self.eg_banner = "<i>not supported</i>"
        self.eg_episode_thumbnails = "<i>not supported</i>"
        self.eg_season_posters = "<i>not supported</i>"
        self.eg_season_banners = "<i>not supported</i>"
        self.eg_season_all_poster = "<i>not supported</i>"
        self.eg_season_all_banner = "<i>not supported</i>"

    # Override with empty methods for unsupported features
    def retrieveShowMetadata(self, folder):
        # no show metadata generated, we abort this lookup function
        return None, None, None

    def create_show_metadata(self, show_obj: "TVShow"):
        pass

    def update_episode_metadata(self, episode_object: "TVEpisode"):
        if self.episode_metadata and episode_object:
            logger.debug(f"[{self.name} META] Updating episode metadata for {episode_object.pretty_name}")
            return self.write_ep_file(episode_object)
        return False

    def update_show_indexer_metadata(self, show_obj: "TVShow"):
        pass

    def get_show_file_path(self, show_obj: "TVShow"):
        pass

    def create_fanart(self, show_obj: "TVShow"):
        pass

    def create_poster(self, show_obj: "TVShow"):
        pass

    def create_banner(self, show_obj: "TVShow"):
        pass

    def create_episode_thumb(self, episode_object: "TVEpisode"):
        pass

    @staticmethod
    def get_episode_thumb_path(episode_object: "TVEpisode"):
        pass

    def create_season_posters(self, episode_object: "TVEpisode"):
        pass

    def create_season_banners(self, episode_object: "TVEpisode"):
        pass

    def create_season_all_poster(self, show_obj: "TVShow"):
        pass

    def create_season_all_banner(self, show_obj: "TVShow"):
        pass

    # Override generic class
    def get_episode_file_path(self, episode_object: "TVEpisode"):
        """
        Returns a full show dir/.meta/episode.txt path for Tivo
        episode metadata files.

        Note, that pyTivo requires the metadata filename to include the original extention.

        ie If the episode name is foo.avi, the metadata name is foo.avi.txt

        episode_object: a TVEpisode object to get the path for
        """
        if os.path.isfile(episode_object.location):
            metadata_filename = os.path.basename(episode_object.location) + "." + self._ep_nfo_extension
            metadata_dir_name = os.path.join(os.path.dirname(episode_object.location), ".meta")
            metadata_file_path = os.path.join(metadata_dir_name, metadata_filename)
        else:
            logger.debug(f"[{self.name} META] Episode location doesn't exist: {episode_object.location}")
            return ""
        return metadata_file_path

    def episode_pretty_title(self, episode_object: "TVEpisode"):
        """
        Returns the name of this episode in a "pretty" human-readable format

        Returns: A string representing the episode's name and season/ep numbers
        """

        if episode_object.show.anime and not episode_object.show.scene:
            return episode_object.naming_pattern("%AB - %EN")
        elif episode_object.show.air_by_date:
            return episode_object.naming_pattern("%AD - %EN")

        return episode_object.naming_pattern("S%0SE%0E - %EN")

    def _ep_data(self, episode_object: "TVEpisode"):
        """
        Creates a key value structure for a Tivo episode metadata file and
        returns the resulting data object.

        episode_object: a TVEpisode instance to create the metadata file for.

        The key values for the tivo metadata file are from:

        https://pytivo.sourceforge.net/wiki/index.php/Metadata
        """

        data = []

        eps_to_write = [episode_object] + episode_object.related_episodes

        indexer_show = episode_object.idxr.series_from_episode(episode_object)
        if not indexer_show:
            logger.debug("Unable to connect to {} while creating meta files for {}, skipping".format(episode_object.indexer_name, episode_object.name))
            return False

        for curEpToWrite in eps_to_write:
            indexer_episode = curEpToWrite.idxr.episode(curEpToWrite)
            if not indexer_episode:
                logger.info(
                    "Metadata writer is unable to find episode {0:d}x{1:d} of {2} on {3}...has it been removed? Should I delete from db?".format(
                        curEpToWrite.season, curEpToWrite.episode, curEpToWrite.show.name, episode_object.indexer_name
                    )
                )
                return False

            if episode_object.airdate != datetime.date.min and not indexer_episode.get("firstAired"):
                indexer_episode["firstAired"] = str(episode_object.airdate)

            if not (indexer_episode.get("episodeName") and indexer_episode.get("firstAired")):
                return None

            if indexer_show.seriesName:
                data.append(f"title : {indexer_show.seriesName}")
                data.append(f"seriesTitle : {indexer_show.seriesName}")

            # noinspection PyProtectedMember
            data.append(f"episodeTitle : {self.episode_pretty_title(curEpToWrite)}")

            # This should be entered for episodic shows and omitted for movies. The standard tivo format is to enter
            # the season number followed by the episode number for that season. For example, enter 201 for season 2
            # episode 01.

            # This only shows up if you go into the Details from the Program screen.

            # This seems to disappear once the video is transferred to TiVo.

            # NOTE: May not be correct format, missing season, but based on description from wiki leaving as is.
            data.append(f"episodeNumber : {curEpToWrite.episode}")

            # Must be entered as true or false. If true, the year from originalAirDate will be shown in parentheses
            # after the episode's title and before the description on the Program screen.

            # FIXME: Hardcode isEpisode to true for now, not sure how to handle movies
            data.append("isEpisode : true")

            # Write the synopsis of the video here
            # Micrsoft Word's smartquotes can die in a fire.
            sanitizedDescription = curEpToWrite.description
            # Replace double curly quotes
            sanitizedDescription = sanitizedDescription.replace("\u201c", '"').replace("\u201d", '"')
            # Replace single curly quotes
            sanitizedDescription = sanitizedDescription.replace("\u2018", "'").replace("\u2019", "'").replace("\u02BC", "'")

            data.append(f"description : {sanitizedDescription}")

            # Usually starts with "SH" and followed by 6-8 digits.
            # Tivo uses zap2it for thier data, so the series id is the zap2itId.
            if getattr(indexer_show, "zap2itId", None):
                data.append(f"seriesId : {indexer_show.zap2itId}")

            # This is the call sign of the channel the episode was recorded from.
            if getattr(indexer_show, "network", None):
                data.append(f"callsign : {indexer_show.network}")

            # This must be entered as yyyy-mm-ddThh:mm:ssZ (the t is capitalized and never changes, the Z is also
            # capitalized and never changes). This is the original air date of the episode.
            # NOTE: Hard coded the time to T00:00:00Z as we really don't know when during the day the first run happened.
            if curEpToWrite.airdate != datetime.date.min:
                data.append(f"originalAirDate : {curEpToWrite.airdate}T00:00:00Z")

            # This shows up at the beginning of the description on the Program screen and on the Details screen.
            for actor in episode_object.idxr.actors(indexer_show):
                actor_name = actor.get("name", "").strip()
                if actor_name:
                    data.append(f"vActor : {actor_name}")

            # This is shown on both the Program screen and the Details screen.
            if indexer_episode.get("siteRating"):
                try:
                    rating = float(indexer_episode["siteRating"])
                except ValueError:
                    rating = 0.0
                # convert 10 to 4 star rating. 4 * rating / 10
                # only whole numbers or half numbers work. multiply by 2, round, divide by 2.0
                rating = round(8 * rating / 10) / 2.0
                data.append(f"starRating : {rating}")

            # This is shown on both the Program screen and the Details screen.
            # It uses the standard TV rating system of: TV-Y7, TV-Y, TV-G, TV-PG, TV-14, TV-MA and TV-NR.
            if getattr(indexer_show, "rating", None):
                data.append(f"tvRating : {indexer_show.rating}")

            # This field can be repeated as many times as necessary or omitted completely.
            if episode_object.show.genre:
                for genre in episode_object.show.genre:
                    if genre:
                        data.append(f"vProgramGenre : {genre}")

                        # NOTE: The following are metadata keywords are not used
                        # displayMajorNumber
                        # showingBits
                        # displayMinorNumber
                        # colorCode
                        # vSeriesGenre
                        # vGuestStar, vDirector, vExecProducer, vProducer, vWriter, vHost, vChoreographer
                        # partCount
                        # partIndex

        return "\n".join(data)

    def write_ep_file(self, episode_object: "TVEpisode"):
        """
        Generates and writes episode_object's metadata under the given path with the
        given filename root. Uses the episode's name with the extension in
        _ep_nfo_extension.

        episode_object: TVEpisode object for which to create the metadata

        filename_path: The file name to use for this metadata. Note that the extension
                will be automatically added based on _ep_nfo_extension. This should
                include an absolute path.
        """
        data = self._ep_data(episode_object)

        if not data:
            return False

        nfo_file_path = self.get_episode_file_path(episode_object)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                logger.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                helpers.chmodAsParent(nfo_file_dir)

            logger.debug("Writing episode nfo file to " + nfo_file_path)

            with open(nfo_file_path, "w") as nfo_file:
                nfo_file.write(data)

            helpers.chmodAsParent(nfo_file_path)

        except EnvironmentError as error:
            logger.error(f"Unable to write file to {nfo_file_path} - are you sure the folder is writable? {error}")
            return False

        return True


# present a standard "interface" from the module
metadata_class = TIVOMetadata



import json
import os

from tornado.web import addslash

import sickchill.start
from sickchill import settings
from sickchill.helper import try_int
from sickchill.helper.common import try_float
from sickchill.oldbeard import config, ui
from sickchill.oldbeard.providers.newznab import NewznabProvider
from sickchill.oldbeard.providers.rsstorrent import TorrentRssProvider
from sickchill.providers.GenericProvider import GenericProvider
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from . import Config


@Route("/config/providers(/?.*)", name="config:providers")
class ConfigProviders(Config):
    @addslash
    def index(self):
        t = PageTemplate(rh=self, filename="config_providers.mako")

        return t.render(
            submenu=self.ConfigMenu(),
            title=_("Config - Providers"),
            header=_("Search Providers"),
            topmenu="config",
            controller="config",
            action="providers",
        )

    @staticmethod
    def canAddNewznabProvider(name):
        if not name:
            return json.dumps({"error": "No Provider Name specified"})

        provider_dict = {x.get_id(): x for x in settings.newznab_provider_list}

        provider_id = GenericProvider.make_id(name)

        if provider_id in provider_dict:
            return json.dumps({"error": "Provider Name already exists as " + name})
        else:
            return json.dumps({"success": provider_id})

    @staticmethod
    def getNewznabCategories(name, url, key):
        """
        Retrieves a list of possible categories with category id's
        Using the default url/api?cat
        https://yournewznaburl.com/api?t=caps&apikey=yourapikey
        """
        error = ""

        if not name:
            error += "\n" + _("No Provider Name specified")
        if not url:
            error += "\n" + _("No Provider Url specified")
        if not key:
            error += "\n" + _("No Provider Api key specified")

        if error:
            return json.dumps({"success": False, "error": error})

        # Get newznabprovider obj with provided name
        temp_provider = NewznabProvider(name, url, key)

        success, tv_categories, error = temp_provider.get_newznab_categories()

        return json.dumps({"success": success, "tv_categories": tv_categories, "error": error})

    @staticmethod
    def deleteNewznabProvider(nnid):
        provider_dict = {x.get_id(): x for x in settings.newznab_provider_list}
        if nnid not in provider_dict or provider_dict[nnid].default:
            return "0"

        # delete it from the list
        settings.newznab_provider_list.remove(provider_dict[nnid])

        if nnid in settings.PROVIDER_ORDER:
            settings.PROVIDER_ORDER.remove(nnid)

        return "1"

    @staticmethod
    def canAddTorrentRssProvider(name, url, cookies, titleTAG):
        if not name:
            return json.dumps({"error": "Invalid name specified"})

        url = config.clean_url(url)
        temp_provider = TorrentRssProvider(name, url, cookies, titleTAG)

        if temp_provider.get_id() in (x.get_id() for x in settings.torrent_rss_provider_list):
            return json.dumps({"error": "Exists as " + temp_provider.name})
        else:
            (succ, errMsg) = temp_provider.validateRSS()
            if succ:
                return json.dumps({"success": temp_provider.get_id()})
            else:
                return json.dumps({"error": errMsg})

    @staticmethod
    def deleteTorrentRssProvider(provider_id):
        provider_dict = {x.get_id(): x for x in settings.torrent_rss_provider_list}

        if provider_id not in provider_dict:
            return "0"

        # delete it from the list
        settings.torrent_rss_provider_list.remove(provider_dict[provider_id])

        if provider_id in settings.PROVIDER_ORDER:
            settings.PROVIDER_ORDER.remove(provider_id)

        return "1"

    def saveProviders(self):
        newznab_provider_dict = {x.get_id(): x for x in settings.newznab_provider_list}

        finished_names = []

        newznab_string = self.get_body_argument("newznab_string", default="")
        torrent_rss_string = self.get_body_argument("torrent_rss_string", default="")
        provider_order = self.get_body_argument("provider_order", default="")

        print(provider_order)

        for current_newznab_string in newznab_string.split("!!!"):
            if not current_newznab_string:
                continue

            name, url, key, categories = current_newznab_string.split("|")
            url = config.clean_url(url)
            provider_id = GenericProvider.make_id(name)

            # if it does not already exist then add it
            if provider_id not in newznab_provider_dict:
                new_provider = NewznabProvider(name, url, key=key, categories=categories)
                settings.newznab_provider_list.append(new_provider)
                newznab_provider_dict[provider_id] = new_provider

            # set all params
            newznab_provider_dict[provider_id].name = name
            newznab_provider_dict[provider_id].url = url
            newznab_provider_dict[provider_id].key = key
            newznab_provider_dict[provider_id].categories = categories
            # a 0 in the key spot indicates that no key is needed
            newznab_provider_dict[provider_id].needs_auth = key and key != "0"
            newznab_provider_dict[provider_id].search_mode = self.get_body_argument(provider_id + "_search_mode", "episode")
            newznab_provider_dict[provider_id].search_fallback = config.checkbox_to_value(
                self.get_body_argument(provider_id + "search_fallback", 0), value_on=1, value_off=0
            )
            newznab_provider_dict[provider_id].enable_daily = config.checkbox_to_value(
                self.get_body_argument(provider_id + "enable_daily", 0), value_on=1, value_off=0
            )
            newznab_provider_dict[provider_id].enable_backlog = config.checkbox_to_value(
                self.get_body_argument(provider_id + "enable_backlog", 0), value_on=1, value_off=0
            )

            # mark it finished
            finished_names.append(provider_id)

        # delete anything that is in the list that was not processed just now
        if newznab_string:
            for provider in settings.newznab_provider_list:
                if provider.get_id() not in finished_names:
                    settings.newznab_provider_list.remove(provider)
                    del newznab_provider_dict[provider.get_id()]

        torrent_rss_provider_dict = {x.get_id(): x for x in settings.torrent_rss_provider_list}

        finished_names = []

        if torrent_rss_string:
            for current_torrent_rss_provider_string in torrent_rss_string.split("!!!"):
                if not current_torrent_rss_provider_string:
                    continue

                name, url, current_cookies, current_title_tag = current_torrent_rss_provider_string.split("|")
                url = config.clean_url(url)
                provider_id = GenericProvider.make_id(name)

                # if it does not already exist then create it
                if provider_id not in torrent_rss_provider_dict:
                    new_provider = TorrentRssProvider(name, url, current_cookies, current_title_tag)
                    settings.torrent_rss_provider_list.append(new_provider)
                    torrent_rss_provider_dict[provider_id] = new_provider

                # update values
                torrent_rss_provider_dict[provider_id].name = name
                torrent_rss_provider_dict[provider_id].url = url
                torrent_rss_provider_dict[provider_id].cookies = current_cookies
                torrent_rss_provider_dict[provider_id].current_title_tag = current_title_tag

                # mark it finished
                finished_names.append(provider_id)

        # delete anything that is in the list that was not processed just now
        if torrent_rss_string:
            for provider in settings.torrent_rss_provider_list:
                if provider.get_id() not in finished_names:
                    settings.torrent_rss_provider_list.remove(provider)
                    del torrent_rss_provider_dict[provider.get_id()]

        # do the enable/disable
        enabled_provider_list = []
        disabled_provider_list = []

        for provider_id, enabled in (provider.split(":") for provider in provider_order.split()):
            enabled = bool(try_int(enabled))

            current_provider_object = [x for x in sickchill.oldbeard.providers.sorted_provider_list() if x.get_id() == provider_id and hasattr(x, "enabled")]

            if current_provider_object:
                current_provider_object[0].enabled = enabled

            if enabled:
                enabled_provider_list.append(provider_id)
            else:
                disabled_provider_list.append(provider_id)

            if provider_id in newznab_provider_dict:
                newznab_provider_dict[provider_id].enabled = enabled
            elif provider_id in torrent_rss_provider_dict:
                torrent_rss_provider_dict[provider_id].enabled = enabled

        # dynamically load provider settings
        for provider in sickchill.oldbeard.providers.sorted_provider_list():
            provider.check_set_option(self, "custom_url")
            provider.check_set_option(self, "cookies")

            provider.check_set_option(self, "minseed", 0, int)
            provider.check_set_option(self, "minleech", 0, int)
            provider.check_set_option(self, "cat", 0, int)
            provider.check_set_option(self, "digest", None)
            provider.check_set_option(self, "hash", None)
            provider.check_set_option(self, "api_key", None, unhide=True)
            provider.check_set_option(self, "username", None, unhide=True)
            provider.check_set_option(self, "password", None, unhide=True)
            provider.check_set_option(self, "passkey", None, unhide=True)
            provider.check_set_option(self, "pin", None, unhide=True)

            provider.check_set_option(self, "confirmed", False, cast=config.checkbox_to_value)
            provider.check_set_option(self, "ranked", False, cast=config.checkbox_to_value)
            provider.check_set_option(self, "engrelease", False, cast=config.checkbox_to_value)
            provider.check_set_option(self, "only_spanish_search", False, cast=config.checkbox_to_value)
            provider.check_set_option(self, "freeleech", False, cast=config.checkbox_to_value)
            provider.check_set_option(self, "search_fallback", False, cast=config.checkbox_to_value)
            provider.check_set_option(self, "enable_daily", False, cast=config.checkbox_to_value)  # fix can_daily
            provider.check_set_option(self, "enable_backlog", False, cast=config.checkbox_to_value)  # fix can_backlog
            provider.check_set_option(self, "subtitle", False, cast=config.checkbox_to_value)

            provider.check_set_option(self, "sorting", "seeders")
            provider.check_set_option(self, "search_mode", "episode")

            provider.check_set_option(self, "ratio", 0, cast=lambda x: max(try_float(x), -1))

        settings.NEWZNAB_DATA = "!!!".join([x.config_string() for x in settings.newznab_provider_list])
        settings.PROVIDER_ORDER = enabled_provider_list + disabled_provider_list

        sickchill.start.save_config()

        # Add a site_message if no providers are enabled for daily and/or backlog
        sickchill.oldbeard.providers.check_enabled_providers()

        ui.notifications.message(_("Configuration Saved"), os.path.join(settings.CONFIG_FILE))

        return self.redirect("/config/providers/")
